/*
 *	The BSL Compiler Kit
 *	A Portable BSL-Compiler, C-Version
 *	Copyright (C) 1995 Nils M. Holm
 *
 *	THE DECLARATION PARSER
 *
 *	Parses declaration statements, and builds the symbol table.
 *
 *	Public functions:
 *		FINDSYM()	Find a symbol in the symbol table
 *		CLRLOCALS()	Clear local symbols
 *		DECLARE()	Parse a declaration
 *		FNDECL()	Parse a function declaration
 *		FWDDECL()	Parse a FORWARD declaration
 *		STCDECL()	Parse a STRUCTURE declaration
 *		STCSIZE()	Determine the size of a structure
 *		SIZE_OF()	Determine the size of an object
 */

#include <stdio.h>
#include <string.h>
#include "defs.h"
#include "data.h"
#include "error.h"
#include "token.h"
#include "glue.h"

#include "proto.h"


char	*strsave();
struct node	*newnode();
struct node	*expr();
struct node	*link();
struct node	*cut();


/*
 *	SY_INIT()
 *	Purpose:	Initialize the symbol table
 *	Receives:	NSYM	- The number of symbol table entries
 *				  to allocate
 *	Returns:	-
 */
void sy_init(nsym) {
	if ((Symbols = (struct syment *) malloc(nsym*sizeof(struct syment)))
	== NULL)
		error(EMEM, FATAL);
	Stsize = nsym;
	St0 = 0;
	St1 = Stsize-1;
	Forwp = 0;
}


/*
 *	NEWGLOB()
 *	Purpose:	Allocates an entry in the global symbol table.
 *	Receives:	-
 *	Returns:	A pointer to a free entry
 */
struct syment *newglob() {
	if (St0 >= St1) error(ESYMENT, FATAL);

	return(&Symbols[St0++]);
}


/*
 *	NEWLOCL()
 *	Purpose:	Allocates an entry in the local symbol table.
 *	Receives:	-
 *	Returns:	A pointer to a free entry
 */
struct syment *newlocl() {
	if (St0 >= St1) error(ESYMENT, FATAL);

	return(&Symbols[St1--]);
}


/*
 *	FINDSYM(NAME, GLOB)
 *	Purpose:	Finds a symbol in the symbol table.
 *	Receives:	NAME - Name of symbol to search
 *			GLOB - Search global symbol table
 *	Returns:	A pointer to the entry found, or NULL,
 *			when the symbol does not exist.
 */
struct syment *findsym(name, glob)
char	*name;
int	glob;
{
	int	si;
	int	s, e;

	s = glob? 0: St1+1;
	e = glob? St0: Stsize;

	for (si=s; si < e; si++) {
		if (*Symbols[si].name == *name &&
		!strcmp(Symbols[si].name, name))
			return(&Symbols[si]);
	}

	return(NULL);
}


/*
 *	ADDGLOB(NAME, FLAGS, SIZE, STDECL)
 *	Purpose:	Adds a global symbol to the symbol table.
 *			When STDECL is FALSE, ADDGLOB() checks for
 *			duplicate globals.
 *	Receives:	NAME	- the symbol name
 *			FLAGS	- the symbol flags (see DEFS.H)
 *			SIZE	- the size of the associated object.
 *			STDECL	- the new symbol is part of a
 *				structure definition.
 *	Returns:	A pointer to the new entry.
 */
struct syment *addglob(name, flags, size)
char	*name;
uint	flags, size;
{
	struct syment	*new;

	if (findsym(name, 1) != NULL) {
		error(EDGLOB, ERROR); return(NULL);
	}

	if (*name == *START && !strcmp(name, START)) flags |= PUBLIC;

	new = newglob();
	new->name = strsave(name);
	new->addr = Dtop;
	new->size = size;
	new->flags = flags;
	new->level = 0;
	return(new);
}


/*
 *	ADDLOCL(NAME, FLAGS, SIZE)
 *	Purpose:	Adds a new symbol to the local symbol table.
 *			ADDLOCL() check for duplicate locals on the
 *			same level, and shadowing of global symbols.
 *	Receives:	NAME	- the symbol name
 *			FLAGS	- the symbol flags (see DEFS.H)
 *			SIZE	- the size of the associated object.
 *	Returns:	A pointer to the new entry.
 */
struct syment *addlocl(name, flags, size)
char	*name;
uint	flags, size;
{
	struct syment	*new, *old;
	int	levl = Level<1? 1: Level;

	if ((old = findsym(name, 0)) != NULL) {
		if (old->level == levl) {
			error(EDLOCL, ERROR); return(NULL);
		}
	}
	else if (findsym(name, 1)) error(ESHAD, WARNING);

	new = newlocl();
	new->name = strsave(name);
	new->addr = (flags&STATIC)? Dtop: Ftop;
	new->size = size;
	new->flags = flags;
	new->level = levl;
	return(new);
}


/*
 *	CLRLOCALS(N)
 *	Purpose:	Clear local symbols of all levels down to N.
 *	Receives:	N	- The level to clear
 *	Returns:	yes.
 */
void clrlocals(n) {
	while (St1 < Stsize-1 && Symbols[St1+1].level >= n) {
		if (!(Symbols[++St1].flags & REFERD))
			unreferenced(Symbols+St1, "local");
		free(Symbols[St1].name);
	}
}


/*
 *	VECSIZE()
 *	Purpose:	Parse the `size' part of a vector declaration.
 *	Receives:	-
 *	Returns:	the size of the vector.
 */
int vecsize() {
	int	size;

	if ((size = cexpr()) < 1) {
		error(ESIZE, ERROR);
		size = 1;
	}
	return(size);
}


/*
 *	ALLOC(SIZE, GLOBL)
 *	Purpose:	Allocates static memory/stack space.
 *	Receives:	SIZE	- number of words to allocate
 *			GLOBL	- allocate global/local space
 *	Returns:	-
 */
void alloc(size, globl) {
	if (globl) {
		if ((Dtop += size) >= DSMAX) error(EDSOVF, FATAL);
	}
	else {
		if ((Ftop += size) >= STKLEN) error(ESTKOVF, WARNING);
	}
}


/*
 *	D_INIT(T, V)
 *	Purpose:	Allocates one word, and emit code to initialize it.
 *	Receives:	TYPE	- type (CODE, DATA)
 *			V	- initial value
 *	Returns:	-
 */
struct node *d_init(type, v) {
	struct node	*t;

	t = newnode(G_INI);
	t->v.val = 1;

	t = prepend(t, G_NUM);
	t->v.val = Dtop;

	if (type != NONE) {
		t = prepend(t, type==CODE? G_RLA: G_REL);
	}

	t = prepend(t, G_NUM);
	t->v.val = v;

	alloc(1, 1);
	return(t);
}


/*
 *	GLOB_INIT(SYM)
 *	Purpose:	Parses the initialization of a global symbol.
 *	Receives:	SYM	- A pointer to the symbol table entry.
 *	Returns:	-
 */
struct node *glob_init(sym)
struct syment	*sym;
{
	struct node	*t;
	int	addr;

	if (Token == LBRACK) {
		next();
		t = mktable(&addr);
		t = link(t, d_init(DATA, addr));
		sym->addr = Dtop-1;
	}
	else if (Token == STRING) {
		t = d_init(DATA, Dtop+1);
		t = link(t, mkstring());
	}
	else {
		t = d_init(NONE, cexpr());
	}
	return(t);
}


/*
 *	LOCL_INIT(STATC)
 *	Purpose:	Parses the initialization of a local symbol.
 *	Receives:	STATC	- the symbol is STATIC
 *	Returns:	A parse tree containing the required
 *			instructions.
 */
struct node *locl_init(statc) {
	struct node	*root;

	if (statc) {
		root = newnode(G_SG);
		root->v.val = Dtop;
		root->right = newnode(G_DRO);
	}
	else {
		root = newnode(G_PSH);
	}
	alloc(1, statc);
	root->left = expr(0);
	return(root);
}


/*
 *	GLOB_ALLOC(SIZE, VEC)
 *	Purpose:	Allocates space for a global variable, or a vector.
 *	Receives:	SIZE	- vector size
 *			VEC	- the space is for a vector
 *	Returns:	-
 */
struct node *glob_alloc(size, vec) {
	struct node	*t;

	t = vec? d_init(DATA, Dtop+1): NULL;
	alloc(size, 1);
	return(t);
}


/*
 *	LOCL_ALLOC(SIZE, VEC, STATC)
 *	Purpose:	Allocates space for a local variable, or a vector.
 *	Receives:	SIZE	- vector size
 *			VEC	- the space is for a vector
 *			STATC	- the associated symbol is STATIC
 *	Returns:	A parse tree that contains the instructions for
 *			the initialization.
 */
struct node *locl_alloc(size, vec, statc) {
	struct node	*root, *t;

	if (statc) return(glob_alloc(size, vec));

	if (!vec) {
		root = newnode(G_PSH);
		t = newnode(G_LN);
		t->v.val = 0;
		root->left = t;
		alloc(1, 0);
	}
	else {
		root = newnode(G_VEC);
		root->v.val = size;
		root->left = newnode(G_ADJ);
		root->right = newnode(G_PSH);

		t = root->left;
		t->left = newnode(G_LN);
		t = t->left;
		t->v.val = Ftop+size;

		alloc(size+1, 0);
	}

	return(root);
}


/*
 *	DECLARE(GLOBL, FLAGS)
 *	Purpose:	Parse the symbol list of a declaration statement,
 *			and adds symbols to the symbol table.
 *	Receives:	GLOBL	- declaration is in global context
 *			FLAGS	- flags for the new symbols
 *	Returns:	global context:	NULL
 *			local context:	A parse tree that contains
 *			the instructions needed for the initialization.
 */
struct node *declare(globl, flags) {
	struct node	*root = NULL, *t;
	struct syment	*new;
	char	sname[NAMELEN+1];
	int	size, vec;

	if (globl) flags |= STATIC;

	for (;;) {
		vec = 0;
		strncpy(sname, Strval, NAMELEN);
		sname[NAMELEN] = 0;
		symname();

		if (Token == LV && !(flags & RVALUE)) {
			next();
			size = vecsize();
			vec = 1;
		}
		else {
			size = 1;
			vec = 0;
		}

		if (globl)
			new = addglob(sname, flags, size);
		else
			new = addlocl(sname, flags, size);

		if (flags & RVALUE) {
			match(MOV, "'='", ERROR);
			new->addr = cexpr();
		}
		else if (Token == MOV) {
			if (size != 1 || flags&EXTERN)
				error(EINIT, ERROR);
			next();
			if (globl)
				t = glob_init(new);
			else
				t = locl_init((flags&STATIC)!=0);
			root = link(root, t);
		}
		else {
			if (!(flags&EXTERN)) {
				if (globl)
					t = glob_alloc(size, vec);
				else
					t = locl_alloc(size, vec,
					(flags&STATIC)!=0);
				root = link(root, t);
			}
		}

		if (Token != COMMA) break;
		next();
	}

	root = cut(root);
	semi();

	return(root);
}


/*
 *	PARAMETERS()
 *	Purpose:	Parse the parameter list of a function declaration,
 *			and add the parameter names to the symbol table.
 *	Receives:	-
 *	Returns:	The number of parameters
 */
int parameters() {
	int	locals = St1;
	int	i, j, nparm = 0;
	char	sname[NAMELEN+1];

	lparen();
	if (Token != RPAREN) for (;; nparm++) {
		strncpy(sname, Strval, NAMELEN);
		sname[NAMELEN] = 0;
		symname();
		addlocl(sname, 0, 1);
		if (Token != COMMA) break;
		next();
	}
	rparen();
	for (j=nparm+1, i=St1+1; i<=locals; i++, j--) {
		Symbols[i].addr = -j-1;
	}
	return(nparm);
}


/*
 *	FNDECL(FLAGS)
 *	Purpose:	Parses a function declaration, and add the
 *			function name to the symbol table.
 *	Receives:	FLAGS	- The flags for the function symbol.
 *	Returns:	-
 */
void fndecl(flags) {
	char	name[NAMELEN+1];
	int	label, forw = 0;
	struct syment	*fn;

	strncpy(name, Strval, NAMELEN);
	name[NAMELEN] = 0;
	symname();
	if ((fn = findsym(name, 1)) != NULL) {
		if (fn->flags & FORWARD) {
			label = fn->size;
			fn->flags &= ~FORWARD;
			fn->flags |= (STATIC|flags);
		}
		else {
			error(EDGLOB, ERROR);
		}
		forw = 1;
	}
	else {
		fn = addglob(name, STATIC|flags, 0);
		label = newlabel();
	}

	parameters();
	function(label, forw);
}


/*
 *	FWDDECL()
 *	Purpose:	Parses the symbol list of a FORWARD declaration,
 *			and adds the symbols to the symbol table.
 *	Receives:	-
 *	Returns:	-
 */
struct node *fwddecl() {
	char	sname[NAMELEN+1];
	struct syment	*sym;
	struct node	*root = NULL, *t;

	for (;;) {
		strcpy(sname, Strval);
		symname();
		sym = addglob(sname, FORWARD|STATIC, 0);
		if (Forwp >= MAXFORW) error(ENFORW, FATAL);
		t = d_init(CODE, sym->size = newlabel());
		Forwtab[Forwp++] = sym->size;
		if (!root)
			root = t;
		else
			root = link(root, t);
		if (Token != COMMA) break;
		next();
	}
	semi();
	return(root);
}


/*
 *	STCDECL()
 *	Purpose:	Parses the body of a STRUCTURE declaration,
 *			and adds the struture with all its members to
 *			the symbol table.
 *	Receives:	-
 *	Returns:	-
 */
void stcdecl() {
	char	sname[NAMELEN+1];
	struct syment	*base, *mem;
	int	size, offset = 0;

	strcpy(sname, Strval);
	symname();
	if (findsym(sname, 1)) error(EDGLOB, ERROR);
	base = addglob(sname, STRUCT|RVALUE, 0);
	lbrace();

	for (;;) {
		strcpy(sname, Strval);
		symname();
		if (findsym(sname, 1)) error(EDGLOB, ERROR);
		if (Token == LV) {
			next();
			size = vecsize();
		}
		else
			size = 1;
		mem = addglob(sname, RVALUE, size);
		mem->addr = offset;
		offset += size;
		base->size++;
		if (Token != COMMA) break;
		next();
	}

	rbrace();
	semi();
}


/*
 *	STCSIZE(STC)
 *	Purpose:	Computes the size of a structure.
 *	Receives:	A pointer to the symbol table entry of the
 *			structure.
 *	Returns:	the size.
 */
stcsize(stc)
struct syment	*stc;
{
	int	i, size = 0;

	for (i=1; i <= stc->size; i++)
		size += stc[i].size;
	return(size);
}
