/*
 *	The BSL Compiler Kit
 *	A Portable BSL-Compiler, C-Version
 *	Copyright (C) 1995 Nils M. Holm
 *
 *	THE EMITTER
 *
 *	Converts an extended parse tree into a BCODE stream, and writes
 *	it to the output file.
 *
 *	Public functions:
 *		TOK2STR()	Convert token to mnemonic
 *		DO_EMIT()	Emit a string/flush ouput
 *		T_EMIT()	Emit a token
 *		N_EMIT()	Emit a numeric literal
 *		SETLABEL()	Place a label
 *		EMIT()		Emit a complete parse tree
 */

#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "defs.h"
#include "data.h"
#include "error.h"
#include "glue.h"

#include "proto.h"


static char	Ebuf[OBUFLEN];			/* Emitter buffer */
static int	Ep;				/* Emit buffer pointer */
static int	Lc;				/* Emitter: chars/line */


struct node	*optimize();


/*
 *	EM_INIT()
 *	Purpose:	Initialize the emitter.
 *	Receives:	FNAME	- The name of the output file
 *	Returns:	-
 */
void em_init(fname)
char	*fname;
{
	Ep = Lc = 0;
	if (fname == NULL) {
		Outfile = fileno(stdout);
		Outname = "(stdout)";
	}
	else {
		if ((Outfile = creat(fname, S_IREAD|S_IWRITE)) < 0)
			error(EINPUT, FATAL);
	}
	setmode(Outfile, O_BINARY);
}


/*
 *	TOK2STR(T)
 *	Purpose:	Convert a token into a BCODE mnemonic.
 *	Receives:	T	- a token.
 *	Returns:	A pointer to the mnemonic string.
 */
char *tok2str(t) {
	switch (t) {
		case G_ADD:	return("ADD");
		case G_SUB:	return("SUB");
		case G_MUL:	return("MUL");
		case G_DIV:	return("DIV");
		case G_MOD:	return("MOD");
		case G_NEG:	return("NEG");

		case G_AND:	return("AND");
		case G_OR:	return("OR");
		case G_XOR:	return("XOR");
		case G_SHL:	return("SHL");
		case G_SHR:	return("SHR");
		case G_ROL:	return("ROL");
		case G_ROR:	return("ROR");
		case G_COM:	return("COM");

		case G_EQ:	return("EQ");
		case G_NE:	return("NE");
		case G_LT:	return("LT");
		case G_GT:	return("GT");
		case G_LE:	return("LE");
		case G_GE:	return("GE");
		case G_ULT:	return("ULT");
		case G_UGT:	return("UGT");
		case G_ULE:	return("ULE");
		case G_UGE:	return("UGE");
		case G_NOT:	return("NOT");

		case G_LN:	return("LN");
		case G_LA:	return("LA");
		case G_LC:	return("LC");
		case G_LL:	return("LL");
		case G_LG:	return("LG");
		case G_LX:	return("LX");
		case G_LV:	return("LV");
		case G_SL:	return("SL");
		case G_SG:	return("SG");
		case G_SX:	return("SX");
		case G_SV:	return("SV");
		case G_ADJ:	return("ADJ");
		case G_SAV:	return("SAV");

		case G_LBL:	return("LBL");
		case G_JMP:	return("JMP");
		case G_JFA:	return("JFA");
		case G_JTR:	return("JTR");
		case G_VJP:	return("VJP");
		case G_SWI:	return("SWI");
		case G_HLT:	return("HLT");

		case G_VCA:	return("VCA");
		case G_FRM:	return("FRM");
		case G_RTN:	return("RTN");
		case G_PSH:	return("PSH");
		case G_RMA:	return("RMA");
		case G_VEC:	return("VEC");
		case G_UNV:	return("UNV");

		case G_INI:	return("INI");
		case G_REL:	return("R");
		case G_RLA:	return("L");
		case G_PUB:	return("PUB");
		case G_EXT:	return("EXT");
		case G_END:	return("END");
		case G_NAM:	return("NAM");

		case G_DRO:	return("DRO");

		case G_LIN:	return("LIN");

		default:	internal(1);
				return(0);
	}
}


/*
 *	DO_EMIT(S)
 *	Purpose:	Emit a string, or flush the output buffer.
 *			S==NULL flushes the buffer.
 *	Receives:	S	- the string to emit, or NULL
 *	Returns:	-
 */
void do_emit(s)
char	*s;
{
	int	l;

	if (O_debug) return;

	if (!s) {
		do_emit(NEWLINE);
		write(Outfile, Ebuf, Ep);
		Ep = 0;
		return;
	}

	l = strlen(s);
	if (l+Ep+3 >= OBUFLEN) {
		write(Outfile, Ebuf, Ep);
		Ep = 0;
	}
	strcpy(Ebuf+Ep, s);
	Ep += l;
	if ((Lc += (l+1)) > 66) {
		Lc = 0;
		Ebuf[Ep] = 0;
		strcat(Ebuf, NEWLINE);
		Ep += (sizeof(NEWLINE)-1);
	}
	else {
		Ebuf[Ep++] = ' ';
	}
}


/*
 *	N_EMIT(N)
 *	Purpose:	Emit a numeric value
 *	Receives:	N	- value to emit
 *	Returns:	-
 */
void n_emit(n) {
	char	buf[12];

	sprintf(buf, "%d", n);
	do_emit(buf);
}


/*
 *	T_EMIT(T)
 *	Purpose:	Emit a token
 *	Receives:	T	- token to emit
 *	Returns:	-
 */
void t_emit(t) {
	do_emit(tok2str(t));
}


/*
 *	EMITNODE(NODE)
 *	Purpose:	Emit a node, and its parameter (if any).
 *			Ignore GLUE nodes.
 *	Receives:	NODE	- A pointer to the node to emit.
 *	Returns:	-
 */
void emitnode(node)
struct node	*node;
{
	if (node->token == GLUE) return;

	if (node->token != G_NUM) do_emit(tok2str(node->token));

	switch (node->token) {
		case G_LN: case G_LA: case G_LL: case G_LG: case G_LX:
		case G_SL: case G_SG: case G_SX: case G_LBL:
		case G_JMP: case G_JFA: case G_JTR: case G_SWI:
		case G_RMA: case G_VEC: case G_UNV: case G_NUM:
		case G_INI: case G_LIN:

			n_emit(node->v.val);
	}
}


/*
 *	EMITTREE(ROOT)
 *	Purpose:	Emit a parse tree in postfix order.
 *	Receives:	ROOT	- A pointer to the root node.
 *	Returns:	-
 */
void emittree(root)
struct node	*root;
{
	if (!root) return;
	emittree(root->left);
	emittree(root->right);
	emitnode(root);
}


/*
 *	EMIT(ROOT)
 *	Purpose:	Emit a parse tree, invoke optimizer/debugger
 *	Receives:	ROOT	- A pointer to the root node.
 *	Returns:	-
 */
void emit(root)
struct node	*root;
{
	if (O_opt) root = optimize(root);
	if (O_debug) dump(root);
	emittree(root);
}
