/*
 *	The BSL Compiler Kit
 *	A Portable BSL-Compiler, C-Version
 *	Copyright (C) 1995 Nils M. Holm
 *
 *	ERROR HANDLER
 *
 *	All ERROR() calls go here.
 *
 *	Public functions:
 *		ERROR()		report an error, and recover
 *		EXPECT()	report a syntax error
 *		INTERNAL()	report a compiler error
 */

#include <stdio.h>
#include <string.h>
#include "defs.h"
#include "data.h"
#include "token.h"
#include "error.h"

#include "proto.h"


static char	*_msg;


/*
 *	RECOVER()
 *	Purpose:	Resync the parser after a SEVERE error.
 *	Receives:	-
 *	Returns:	-
 */
void recover() {
	next();
}


/*
 *	ERROR(ID, CLASS)
 *	Purpose:	Report/handle an error
 *	Receives:	ID	- Error ID (see ERROR.H)
 *			CLASS	- Error class
 *	Returns:	Nothing. In case of a FATAL error,
 *			it does not return at all.
 */
void error(id, class) {
	char	*msg;

	switch (id) {
		case EUNREFD:
		case EUDECL:
		case ESYN:	msg = _msg; break;

		case ESHAD:	msg = "local decl shadows global"; break;
		case ESTKOVF:	msg = "possible stack overflow"; break;
		case EOVFL:	msg = "numeric literal overflow"; break;
		case ETERM:	msg = "unterminated character literal"; break;
		case ELVAL:	msg = "bad lvalue"; break;
		case EDECL:	msg = "invalid declaration"; break;
		case EDIV:	msg = "divide by zero"; break;
		case EDGLOB:	msg = "duplicate global"; break;
		case EDLOCL:	msg = "duplicate local"; break;
		case ESIZE:	msg = "bad size"; break;
		case EINIT:	msg = "invalid initialization"; break;
		case EBREAK:	msg = "'loop' or 'break' in wrong context";
				break;
		case EDCASE:	msg = "duplicate case"; break;
		case EXADDR:	msg = "address of external is unknown"; break;
		case EAADDR:	msg = "local address is not constant"; break;
		case EMEM:	msg = "insufficient memory"; break;
		case EIO:	msg = "I/O error"; break;
		case EINPUT:	msg = "cannot open input file"; break;
		case EOUTPUT:	msg = "cannot create output file"; break;
		case ESCAN:	msg = "bad input character"; break;
		case ETOKL:	msg = "token too long"; break;
		case ENODES:	msg = "out of free nodes"; break;
		case ENYI:	msg = "not yet implemented"; break;
		case ELABEL:	msg = "too many labels"; break;
		case ESYMENT:	msg = "symbol table overflow"; break;
		case EDSOVF:	msg = "data segment overflow"; break;
		case ETBLL:	msg = "table too big"; break;
		case ENCASE:	msg = "too many cases"; break;
		case ENFORW:	msg = "too many forward declarations"; break;
		case ECOMP:	msg = "compiler error"; break;

		default:	msg = "unknown error"; class = FATAL; break;
	}

	fprintf(stderr, "%s (%d): %s [%c%d]\n", Inname, Lno, msg, class, id);

	if (class == WARNING) return;

	if (class == SEVERE) recover();

	if (class == FATAL) {
		fprintf(stderr, "terminating.\n");
		exit(2);
	}

	Errcount++;
}


/*
 *	EXPECT(WHAT, CLASS)
 *	Purpose:	Report a syntax error
 *	Receives:	WHAT	- A description of the expected token
 *			CLASS	- The class of the error
 *	Returns:	thru ERROR()
 */
void expect(what, class)
char	*what;
int	class;
{
	char	buf[TOKLEN*2];

	sprintf(buf, "syntax error (%s expected)", what);
	_msg = buf;
	error(ESYN, class);
}


/*
 *	UNDECLARED(NAME)
 *	Purpose:	Report reference to an undeclared symbol
 *	Receives:	NAME	- Name of undeclared symbol
 *	Returns:	A pointer to the previously undeclared symbol.
 */
struct syment *undeclared(name)
char	*name;
{
	char	buf[TOKLEN*2];
	struct syment	*fix;

	fix = (Level? addlocl: addglob)(name, Level? 0: STATIC, 1);
	sprintf(buf, "undeclared symbol (%s)", name);
	_msg = buf;
	error(EUDECL, ERROR);
	return(fix);
}


/*
 *	UNREFERENCED(SP, SCOPE)
 *	Purpose:	Report an unused global/local symbol
 *	Receives:	NAME	- Name of undeclared symbol
 *	Returns:	thru ERROR()
 */
void unreferenced(sp, scope)
struct syment	*sp;
char	*scope;
{
	char	buf[TOKLEN*2];

	sprintf(buf, "unreferenced %s symbol: %s", scope, sp->name);
	_msg = buf;
	error(EUNREFD, WARNING);
}


/*
 *	INTERNAL(ID)
 *	Purpose:	Report a compiler error (FATAL)
 *	Receives:	ID	- A unique ID
 *	Returns:	Never, because it causes a FATAL error
 */
void internal(id) {
	printf("Internal error #%d\n", id);
	error(ECOMP, FATAL);
}
