/*
 *	The BSL Compiler Kit
 *	A Portable BSL-Compiler, C-Version
 *	Copyright (C) 1995 Nils M. Holm
 *
 *	THE STATEMENT PARSER
 *
 *	Parses a complete program, and builds/emits an extended parse
 *	tree for each function. Also emits data initialization sequences.
 *
 *	Public functions:
 *		FUNCTION()	parse a function body, and build/emit
 *				the generated code
 *		PARSE()		Parse a program
 */

#include <stdio.h>
#include <string.h>
#include "defs.h"
#include "data.h"
#include "error.h"
#include "scan.h"
#include "token.h"
#include "glue.h"

#include "proto.h"


static int	loop_start,			/* Label at beg. of loop */
		loop_end,			/* Label at end of loop */
		loop_level;			/* Nesting level at */
						/* beginning of loop */


struct node	*newnode();
struct node	*expr();
struct node	*statement();
struct node	*link();
struct node	*cut();
struct node	*declare();
struct node	*declarations();
struct syment	*findsym();


/*
 *	PA_INIT()
 *	Purpose:	Initialize the parser.
 *	Receives:	-
 *	Returns:	-
 */
void pa_init() {
	loop_start = loop_end = BADLABEL;
	Lno = 1;
	Errcount = 0;
	Level = 0;
	Dtop = 0;
	Ftop = 1;
}


/*
 *	IF_STMT()
 *	Purpose:	Parse an IF-ELSE-statement.
 *	Receives:	-
 *	Returns:	An extended parse tree.
 */
struct node *if_stmt() {
	int	true, false;
	struct node	*root, *t;

	next();
	lparen();
	root = newnode(G_LBL);
	root->v.val = false = newlabel();
	t = root->left = newnode(G_JFA);
	t->v.val = false;
	t->right = expr(1);
	rparen();
	root->right = statement();

	if (Token == KELSE) {
		next();
		t = newnode(G_JMP);
		t->v.val = true = newlabel();
		t->right = root->right;
		root->right = t;
		t = newnode(G_LBL);
		t->v.val = true;
		t->left = root;
		root = t;
		root->right = statement();
	}

	return(root);
}


/*
 *	HALT_STMT()
 *	Purpose:	Parse a HALT-statement
 *	Receives:	-
 *	Returns:	An extended parse tree.
 */
struct node *halt_stmt() {
	struct node	*root;

	next();
	root = newnode(G_HLT);
	root->left = expr(0);
	return(root);
}


/*
 *	ADDNODE(ROOT, TYPE)
 *	Purpose:	Add a node to a parse tree. If the head is
 *			a GLUE node, change it, else create a new head.
 *	Receives:	ROOT	- the head of the subtree
 *			TYPE	- the desired token for the head
 *	Returns:	The new head.
 */
struct node *addnode(root, type)
struct node	*root;
int	type;
{
	struct node	*t;

	if (root->token != GLUE) {
		t = newnode(type);
		t->left = root;
		root = t;
	}
	else {
		root->token = type;
	}
	return(root);
}


/*
 *	MKSWITCH(NCASE, DFLT, LBL, VAL)
 *	Purpose:	Build an extended parse tree for a SWITCH-statement.
 *	Receives:	NCASE	- Number of CASEs in the SWITCH
 *			DFLT	- The DEFAULT label
 *			LBL	- A pointer to the label array
 *			VAL	- A pointer to the value array (case values)
 *	Returns:	The parse tree
 */
struct node *mkswitch(ncase, dflt, lbl, val)
int	ncase, dflt, *lbl, *val;
{
	struct node	*root, *t;
	int	i;

	root = newnode(G_SWI);
	root->v.val = ncase;
	t = newnode(G_NUM);
	t->v.val = Dtop;
	t->right = root;
	root = t;
	t = newnode(G_NUM);
	t->v.val = dflt;
	t->right = root;
	root = t;

	for (i=0; i<ncase; i++) {
		t = newnode(G_NUM);
		t->v.val = val[i];
		t->right = root;
		root = t;
		t = newnode(G_NUM);
		t->v.val = lbl[i];
		t->right = root;
		root = t;
	}

	alloc((ncase+1)*2, 1);

	return(root);
}


/*
 *	SWITCH_STMT()
 *	Purpose:	Parse a SWITCH-statement.
 *	Receives:	-
 *	Returns:	An extended parse tree.
 */
struct node *switch_stmt() {
	struct node	*root, *t, *exp;
	int	lbl[MAXCASE],
		val[MAXCASE];
	int	ncase, i;
	int	n_start, n_end, n_level;
	int	dflt = BADLABEL;

	next();
	lparen();
	exp = expr(1);
	rparen();

	root = newnode(G_JMP);
	n_start = loop_start;
	n_end = loop_end;
	n_level = Level;
	loop_start = root->v.val = newlabel();
	loop_end = newlabel();
	loop_level = Ftop;

	lbrace();
	for (ncase=0;;) {
		if (Token == KCASE) {
			if (ncase >= MAXCASE) error(ENCASE, FATAL);
			next();
			root = addnode(root, G_LBL);
			root->token = G_LBL;
			lbl[ncase] = root->v.val = newlabel();
			val[ncase] = cexpr();
			for (i=0; i<ncase; i++) {
				if (val[i] == val[ncase]) {
					error(EDCASE, ERROR);
					break;
				}
			}
			ncase++;
			colon();
		}
		else if (Token == KDFLT) {
			if (dflt != BADLABEL) error(EDCASE, ERROR);
			next();
			colon();
			root = addnode(root, G_LBL);
			root->token = G_LBL;
			dflt = root->v.val = newlabel();
		}
		else if (Token == RBRACE) {
			break;
		}
		else {
			t = statement();
			root = link(root, t);
		}
	}
	rbrace();

	root = addnode(root, G_JMP);
	root->v.val = loop_end;

	root = addnode(root, G_LBL);
	root->v.val = loop_start;
	root = link(root, exp);

	if (dflt == BADLABEL) dflt = loop_end;
	root = link(root, mkswitch(ncase, dflt, lbl, val));
	root->token = G_LBL;
	root->v.val = loop_end;

	loop_start = n_start;
	loop_end = n_end;
	loop_level = n_level;

	return(root);
}


/*
 *	WHILE_STMT()
 *	Purpose:	Parse a WHILE-statement
 *	Receives:	-
 *	Returns:	An extended parse tree.
 */
struct node *while_stmt() {
	struct node	*root, *t;
	int	begin, end, again;
	int	n_start, n_end, n_level;

	next();
	lparen();
	root = newnode(G_JTR);
	root->v.val = again = newlabel();
	t = root->left = newnode(G_LBL);
	root->left->v.val = begin = newlabel();
	root->right = expr(1);
	end = newlabel();
	rparen();

	n_start = loop_start;
	n_end = loop_end;
	n_level = loop_level;
	loop_start = begin;
	loop_end = end;
	loop_level = Ftop;
	t->left= statement();
	loop_start = n_start;
	loop_end = n_end;
	loop_level = n_level;

	t = newnode(G_LBL);
	t->v.val = end;
	t->right = root;
	root = t;
	t = root->left = newnode(G_LBL);
	t->v.val = again;
	t->left = newnode(G_JMP);
	t->left->v.val = begin;

	return(root);
}


/*
 *	UNTIL_STMT()
 *	Purpose:	Parse an UNTIL-statement
 *	Receives:	-
 *	Returns:	An extended parse tree.
 */
struct node *until_stmt() {
	struct node	*root, *t;
	int	begin, end;
	int	n_start, n_end, n_level;

	next();
	root = newnode(G_LBL);
	root->v.val = end = newlabel();
	root->left = newnode(G_LBL);
	root->left->v.val = begin = newlabel();
	t = root->right = newnode(G_JFA);
	t->v.val = begin;

	n_start = loop_start;
	n_end = loop_end;
	n_level = loop_level;
	loop_start = begin;
	loop_end = end;
	loop_level = Ftop;
	t->left = statement();
	loop_start = n_start;
	loop_end = n_end;
	loop_level = n_level;

	match(KUNTIL, "'until'", ERROR);
	lparen();
	t->right = expr(1);
	rparen();


	return(root);
}


/*
 *	FOR_INIT(PSYM)
 *	Purpose:	Parse the INIT part of a FOR-statement:
 *			FOR (VAR=EXPR, EXPR, CEXPR)
 *			     ~~~~~~~~
 *	Receives:	PSYM	- A pointer to the used counting
 *				variable (by reference).
 *	Returns:	A parse tree containing the init code.
 *			A pointer to the used symbol (thru PSYM).
 */
struct node *for_init(psym, begin)
struct syment	**psym;
int	begin;
{
	struct node	*root;
	char	sname[NAMELEN+1];

	strncpy(sname, Strval, NAMELEN);
	sname[NAMELEN] = 0;
	symname();
	if ((*psym = findsym(sname, 0)) == NULL) {
		if ((*psym = findsym(sname, 1)) == NULL) {
			*psym = undeclared(sname);
		}
	}
	(*psym)->flags |= REFERD;
	match(MOV, "'='", ERROR);

	root = newnode(G_JMP);
	root->v.val = begin;
	if ((*psym)->flags & EXTERN) {
		root->right = newnode(G_SX);
		root->right->v.val = findsx((*psym)->name);
	}
	else {
		root->right = newnode(((*psym)->flags & STATIC)? G_SG: G_SL);
		root->right->v.val = (*psym)->addr;
	}
	root->left = expr(0);

	return(root);
}


/*
 *	FOR_COND(SYM, BEGIN, END)
 *	Purpose:	Parse the CONDITION part of a FOR-statement:
 *			FOR (VAR=EXPR, EXPR, CEXPR)
 *				       ~~~~
 *	Receives:	SYM	- A pointer to the symbol table entry of
 *				the used counter.
 *			BEGIN	- The label at the start of the loop
 *			END	- The label at the end of the loop
 *	Returns:	An extended parse tree containing code to
 *			check the condition.
 */
struct node *for_cond(sym, begin, again)
struct syment	*sym;
int	begin, again;
{
	struct node	*root, *t, *t2;

	root = newnode(G_JTR);
	root->v.val = again;
	t = root->right = newnode(GLUE);	/* [U]LT/[U]GT */
	if (sym->flags & EXTERN) {
		t->left = newnode(G_LX);
		t->left->v.val = findsx(sym->name);
	}
	else {
		t->left = newnode((sym->flags & STATIC)? G_LG: G_LL);
		t->left->v.val = sym->addr;
	}
	t2 = t->left;
	t2->left = newnode(G_LBL);
	t2->left->v.val = begin;
	t->right = expr(0);
	return(root);
}


/*
 *	FOR_INCR(SYM, INCR)
 *	Purpose:	Build code for the INCREMENT part of a FOR-statement:
 *			FOR (VAR=EXPR, EXPR, CEXPR)
 *					     ~~~~~
 *	Receives:	SYM	- A pointer to the symbol table entry of
 *				the used counter.
 *			INCR	- the increment (may be negative to do
 *				a decrement).
 *	Returns:	An extended parse tree.
 */
struct node *for_incr(sym, incr, cont)
struct syment	*sym;
int	incr, cont;
{
	struct node	*root, *t;
	int	sx;

	root = newnode(G_ADD);
	if (sym->flags & EXTERN) {
		root->left = newnode(G_LX);
		sx = root->left->v.val = findsx(sym->name);
	}
	else {
		root->left = newnode((sym->flags & STATIC)? G_LG: G_LL);
		root->left->v.val = sym->addr;
	}
	t = root->left;
	t->left = newnode(G_LBL);
	t->left->v.val = cont;
	root->right = newnode(G_LN);
	root->right->v.val = incr;
	if (sym->flags & EXTERN) {
		t = newnode(G_SX);
		t->v.val = sx;
	}
	else {
		t = newnode((sym->flags & STATIC)? G_SG: G_SL);
		t->v.val = sym->addr;
	}
	t->right = root;
	root = newnode(G_DRO);
	root->right = t;
	return(root);
}


/*
 *	FOR_STMT()
 *	Purpose:	Parse a FOR-statement
 *	Receives:	SIGN	- Respect signs in comparisons
 *				(zero for UFOR).
 *	Returns:	An extended parse tree.
 */
struct node *for_stmt(sign) {
	struct node	*root, *t, *op;
	struct syment	*sym;
	int	begin, cont, again, end;
	int	incr;
	int	n_start, n_end, n_level;

	next();
	lparen();
	root = newnode(G_LBL);
	begin = newlabel();
	cont = newlabel();
	again = newlabel();
	root->v.val = end = newlabel();
	root->left = for_init(&sym, begin);
	comma();

	t = root->right = for_cond(sym, begin, again);
	op = t->right;

	if (Token == COMMA) {
		next();
		incr = cexpr();
	}
	else {
		incr = 1;
	}

	t->left = for_incr(sym, incr, cont);
	t = t->left;
	t = t->left = newnode(GLUE);
	t->left = newnode(G_LBL);
	t->left->v.val = again;

	rparen();

	n_start = loop_start;
	n_end = loop_end;
	n_level = loop_level;
	loop_start = cont;
	loop_end = end;
	loop_level = Ftop;
	t->right = statement();
	loop_start = n_start;
	loop_end = n_end;
	loop_level = n_level;

	if (incr < 0) {
		op->token = sign? G_GT: G_UGT;
	}
	else {
		op->token = sign? G_LT: G_ULT;
	}

	return(root);
}


/*
 *	LOOP_STMT(LABEL)
 *	Purpose:	Parse a LOOP/BREAK statement (with sematic analysis).
 *	Receives:	LABEL	- The destination label:
 *				loop_start for LOOP
 *				loop_end for BREAK
 *	Returns:	An extended parse tree.
 */
struct node *loop_stmt(label) {
	struct node	*root;

	next();
	if (label == BADLABEL) {
		error(EBREAK, ERROR);
		return(NULL);
	}
	else {
		root = newnode(G_JMP);
		root->v.val = label;
		root->left = newnode(G_UNV);
		root->left->v.val = Ftop-loop_level;
		return(root);
	}
}


/*
 *	RETURN_STMT()
 *	Purpose:	Parse a RETURN statement
 *	Receives:	-
 *	Returns:	An extended parse tree.
 */
struct node *return_stmt() {
	struct node	*root;

	next();
	root = newnode(G_RTN);
	root->right = newnode(G_UNV);
	root->right->v.val = Ftop-1;
	if (Token != SEMI) root->left = expr(1);
	return(root);
}


/*
 *	SIMPLE()
 *	Purpose:	Parse a simple (non-compound) statement.
 *	Receives:	-
 *	Returns:	An extended parse tree
 */
struct node *simple() {
	struct node	*root;

	switch (Token) {
		case RBRACE:	return (NULL);
		case SEMI:	root = newnode(GLUE); break;
		case KBREAK:	root = loop_stmt(loop_end); break;
		case KDO:	root = until_stmt(); break;
		case KFOR:	root = for_stmt(1); return(root);
		case KUFOR:	root = for_stmt(0); return(root);
		case KHALT:	root = halt_stmt(); break;
		case KIF:	root = if_stmt(); return(root);
		case KLOOP:	root = loop_stmt(loop_start); break;
		case KRETURN:	root = return_stmt(); break;
		case KSWITCH:	root = switch_stmt(); return(root);
		case KWHILE:	root = while_stmt(); return(root);
		default:	root = expr(1);
	}
	semi();
	return(root);
}


/*
 *	COMPOUND()
 *	Purpose:	Increase the nesting level, parse a compound statement,
 *			clear locals, and decrease the level again.
 *	Receives:	-
 *	Returns:	An extended parse tree.
 */
struct node *compound() {
	struct node	*root, *base, *t;
	uint	oftop;

	next();
	++Level;
	oftop = Ftop;
	root = newnode(G_UNV);
	base = newnode(G_DRO);
	base->left = declarations(0);
	for (;;) {
		if ((t = statement()) == NULL) break;
		if (Token == ENDFILE) break;
		base = link(base, t);
	}
	base = cut(base);
	rbrace();
	root->left = base;
	root->v.val = (Ftop - oftop);
	Ftop = oftop;
	clrlocals(Level--);

	return(root);
}


/*
 *	STATEMENT()
 *	Purpose:	Parse a statement (simple or compound).
 *			Signal end of statement for the code generator
 *			(by adding DRO to the parse tree).
 *	Receives:	-
 *	Returns:	An extended parse tree.
 */
struct node *statement() {
	struct node	*root;
	static int	lastln = 0;

	root = newnode(G_DRO);
	switch (Token) {
		case LBRACE:	root->left = compound(); break;
		default:	if (O_linfo && Lno != lastln) {
					root->left = newnode(G_LIN);
					root->left->v.val = Lno;
					lastln = Lno;
					root->right = simple();
				}
				else {
					root->left = simple();
				}
	}
	if (root->left == NULL) {
		freenode(root); root = NULL;
	}
	return(root);
}


/*
 *	FUNCTION()
 *	Purpose:	Parse a function body, emit the generated
 *			code, and clear the node pool.
 *	Receives:	-
 *	Returns:	-
 */
void function(label, forw) {
	struct node	*ptree, *t;

	t = newnode(G_LBL);
	t->v.val = label;
	if (!forw) t->left = d_init(CODE, label);

	ptree = newnode(G_RTN);
	ptree->left = newnode(G_FRM);
	ptree->right = statement();
	ptree = link(t, ptree);
	if (!Errcount) emit(ptree);
	clearnodes();
	clrlocals(0);
}


/*
 *	PARSE_OPTS()
 *	Purpose:	Parse an OPTIONS statement, and set the
 *			requested options.
 *	Receives:	-
 *	Returns:	-
 */
void parse_opts() {
	next();
	if (Token != STRING)
		expect("string literal", ERROR);
	else
		options(Strval);
	next();
	semi();
}


/*
 *	RESYNC()
 *	Purpose:	Re-synchronize line numbers, and change input
 *			file name.
 *	Receives:	-
 *	Returns:	-
 */
void resync() {
	next();
	if (Token != STRING) {
		expect("file name", ERROR);
	}
	else {
		if (*Inname == '(') {
			t_emit(G_NAM);
			do_emit(Strval);
		}
		strcpy(Inname, Strval);
	}
	next();
	if (Token != NUMBER)
		expect("line number", ERROR);
	else
		Lno = Value-1;
	next();
}


/*
 *	DECLARATIONS(GLOBL)
 *	Purpose:	Parse a chain of declarations.
 *	Receives:	GLOBL	- Declarations are in global context.
 *	Returns:	GLOBL	-> NULL
 *			!GLOBL	-> An extended parse tree containing
 *				the allocation/deallocation/initialization
 *				code for the locals.
 */
struct node *declarations(globl) {
	struct node	*root = NULL, *t;

	for (;;) {
		if (Token == KVAR) {
			next();
			t = declare(globl, 0);
		}
		else if (Token == KVALUE) {
			next();
			t = declare(globl, RVALUE);
		}
		else if (Token == KEXTERN) {
			next();
			if (!globl) error(EDECL, ERROR);
			t = declare(globl, EXTERN);
		}
		else if (Token == KSTATIC) {
			next();
			if (Token == KVAR) next();
			t = declare(globl, STATIC);
		}
		else if (Token == KPUBLIC) {
			next();
			if (Token == KVAR) {
				next();
				if (!globl) error(EDECL, ERROR);
				t = declare(1, STATIC|PUBLIC);
			}
			else {
				fndecl(PUBLIC);
				t = NULL;
			}
		}
		else if (Token == KFORWD) {
			next();
			t = fwddecl();
		}
		else if (Token == KSTRUCT) {
			next();
			stcdecl();
			t = NULL;
		}
		else if (Token == KOPTS && globl) {
			parse_opts();
			t = NULL;
		}
		else if (Token == HASH) {
			resync();
			t = NULL;
		}
		else if (globl && Token == SYMBOL) {
			fndecl(0);
			t = NULL;
		}
		else {
			if (Token == ENDFILE || !globl)
				return(globl? NULL: cut(root));
			expect("declaration", SEVERE);
		}

		if (globl) {
			emit(t); clearnodes();
		}
		else {
			root = link(root, t);
		}
	}
}


/*
 *	PARSE()
 *	Purpose:	Parse a complete program.
 *			PARSE() returns only if the EOF has been read
 *			[in DECLARATIONS(TRUE)].
 *	Receives:	-
 *	Returns:	-
 */
void parse() {
	next();
	declarations(1);
}
