/*
 *	YACCable BSL (Basic Systems Language) Grammar, V.1
 *	Copyright (C) 1995 Nils M. Holm
 *
 *	There is one S/R conflict - the well-known dangling else.
 */

%token	BREAK CASE DEFAULT DO ELSE EXTERN IF
%token	FOR FORWARD HALT LOOP OPTIONS PUBLIC
%token	RETURN STRUCTURE SWITCH UFOR
%token	UNSIGNED UNTIL VAR WHILE NUMBER STRING
%token	SYMBOL

%%

/*
 *	---------------   D E C L A R A T I O N S   ---------------
 */

top_level:
		glob_decl_list
	;


/*
 *	GLOBAL DECLARATIONS may only appear at top level.
 */

glob_decl_list:
		glob_decl
	|	glob_decl_list glob_decl
	;

glob_decl:
		glob_var_decl ';'
	|	func_decl
	|	struct_decl ';'
	|	compiler_options ';'
	|	input_sync
	;

compiler_options:
		OPTIONS STRING ';'
	;

input_sync:
		'#' STRING NUMBER
	;

glob_var_decl:
		VAR glob_var_list
	|	PUBLIC VAR glob_var_list
	|	EXTERN simple_var_list
	;

glob_var_list:
		glob_symdecl
	|	glob_var_list ',' glob_symdecl
	;

glob_symdecl:
		SYMBOL
	|	SYMBOL '=' glob_init
	|	SYMBOL '.' const_expr
	;

glob_init:
		const_expr
	|	table
	|	STRING
	;


/*
 *	LOCAL DECLARATIONS may only appear at the beginning
 *	of compound statements.
 */

local_decl_list:
		local_decl
	|	local_decl_list local_decl
	;

local_decl:
		local_var_decl ';'
	|	struct_decl ';'
	;

local_var_decl:
		VAR local_var_list
	;

local_var_list:
		local_symdecl
	|	local_var_list ',' local_symdecl
	;

local_symdecl:
		SYMBOL
	|	SYMBOL '=' simple_expr
	|	SYMBOL '.' const_expr
	;


simple_var_list:
		SYMBOL
	|	SYMBOL ',' simple_var_list
	;


/*
 *	STRUCTURE  DEFINITIONS may appear either at top level,
 *	or at the beginning of compound statements.
 */

struct_decl:
		STRUCTURE '{' member_list '}'
	;

member_list:
		member
	|	member_list ',' member
	;

member:
		SYMBOL
	|	SYMBOL '.' const_expr
	;


/*
 *	FUNCTION DEFINITIONS must be at top level.
 */

func_decl:
		SYMBOL parameters statement
	|	PUBLIC SYMBOL parameters statement
	;

parameters:
		'(' ')'
	|	'(' parameter_list ')'
	;

parameter_list:
		SYMBOL
	|	parameter_list ',' SYMBOL
	;


/*
 *	---------------   S T A T E M E N T S   ---------------
 */

statement:
		simple_stmt
	|	compound_stmt
	;

compound_stmt:
		'{' '}'
	|	'{' local_decl_list '}'
	|	'{' statement_list '}'
	|	'{' local_decl_list statement_list '}'
	;

statement_list:
		statement
	|	statement_list statement
	;

simple_stmt:
		';'
	|	expression ';'
	|	halt_stmt
	|	cond_stmt
	|	loop_stmt
	|	jump_stmt
	;

halt_stmt:
		HALT expression ';'
	;

cond_stmt:
		if_stmt
	|	switch_stmt
	;

loop_stmt:
		do_until_stmt ';'
	|	while_stmt
	|	for_stmt
	;

jump_stmt:
		RETURN ';'
	|	RETURN expression ';'
	|	BREAK ';'
	|	LOOP ';'
	;

if_stmt:
		IF '(' expression ')' statement
	|	IF '(' expression ')' statement
		ELSE statement
	;

switch_stmt:
		SWITCH '(' expression ')' '{'
			switch_body
		'}'
	;

switch_body:
		sw_body_stmt
	|	switch_body sw_body_stmt
	;

sw_body_stmt:
		statement
	|	CASE const_expr ':'
	|	DEFAULT ':'
	;

while_stmt:
		WHILE '(' expression ')' statement
	;

do_until_stmt:
		DO statement UNTIL '(' expression ')'
	;

for_stmt:
		FOR '(' SYMBOL '=' simple_expr ','
		simple_expr opt_increment ')'
			statement
	|	UFOR '(' SYMBOL '=' simple_expr ','
		simple_expr opt_increment ')'
			statement
	;

opt_increment:
		',' const_expr
	|
	;


/*
 *	---------------   E X P R E S S I O N S   ---------------
 */

expression:
		simple_expr
	|	expression ',' simple_expr
	;

simple_expr:
		assignment
	;

assignment:
		conditional
	|	lvalue '=' assignment
	|	lvalue '+=' assignment
	|	lvalue '-=' assignment
	|	lvalue '*=' assignment
	|	lvalue '/=' assignment
	|	lvalue '%=' assignment
	|	lvalue '&=' assignment
	|	lvalue '|=' assignment
	|	lvalue '^=' assignment
	|	lvalue '>>=' assignment
	|	lvalue '<<=' assignment
	|	lvalue '}}=' assignment
	|	lvalue '{{=' assignment
	;

conditional:
		logical_or
	|	logical_or '->' conditional ',' conditional
	;

logical_or:
		logical_and
	|	logical_or '||' logical_and
	;

logical_and:
		binary_or
	|	logical_and '&&' binary_or
	;

binary_or:
		binary_xor
	|	binary_or '|' binary_xor
	;

binary_xor:
		binary_and
	|	binary_xor '^' binary_and
	;

binary_and:
		equation
	|	binary_and '&' equation
	;

equation:
		comparison
	|	equation '==' comparison
	|	equation '!=' comparison
	;

comparison:
		shift_rotate
	|	comparison '>' shift_rotate
	|	comparison '<' shift_rotate
	|	comparison '>=' shift_rotate
	|	comparison '<=' shift_rotate
	|	comparison UNSIGNED '>' shift_rotate
	|	comparison UNSIGNED '<' shift_rotate
	|	comparison UNSIGNED '>=' shift_rotate
	|	comparison UNSIGNED '<=' shift_rotate
	;

shift_rotate:
		sum
	|	shift_rotate '>>' sum
	|	shift_rotate '<<' sum
	|	shift_rotate '}}' sum
	|	shift_rotate '{{' sum
	;

sum:
		term
	|	sum '+' term
	|	sum '-' term
	;

term:
		factor
	|	term '*' factor
	|	term '/' factor
	|	term '%' factor
	;

factor:
		lvalue
	|	'!' factor
	|	'~' factor
	|	'@' primary
	|	'+' factor
	|	'-' factor
	;

lvalue:
		struct_access
	|	'.' factor
	;

struct_access:
		primary opt_fncall
	|	struct_access '.' primary opt_fncall
	;

opt_fncall:
		function_call
	|
	;

function_call:
		'(' ')'
	|	'(' expr_list ')'
	;

expr_list:
		simple_expr
	|	expr_list ',' simple_expr
	;

primary:
		NUMBER
	|	SYMBOL
	|	STRING
	|	'(' expression ')'
	|	table
	;

table:
		'[' table_elements ']'
	;

table_elements:
		table_item
	|	table_elements ',' table_item
	;

table_item:
		'@' SYMBOL
	|	const_expr
	|	table
	;


/*
 *	CONSTANT EXPRESSIONS are used in locations where values
 *	must be known at compile time.
 */

const_expr:
		const_binxor
	|	const_expr '|' const_binxor
	;

const_binxor:
		const_binand
	|	const_binxor '^' const_binand
	;

const_binand:
		const_sh_rot
	|	const_binand '&' const_sh_rot
	;

const_sh_rot:
		const_sum
	|	const_sh_rot '>>' const_sum
	|	const_sh_rot '<<' const_sum
	|	const_sh_rot '}}' const_sum
	|	const_sh_rot '{{' const_sum
	;

const_sum:
		const_term
	|	const_sum '+' const_term
	|	const_sum '-' const_term
	;

const_term:
		const_factor
	|	const_term '*' const_factor
	|	const_term '/' const_factor
	|	const_term '%' const_factor
	;

const_factor:
		const_primary
	|	'!' const_factor
	|	'~' const_factor
	|	'+' const_factor
	|	'-' const_factor
	;

const_primary:
		NUMBER
	|	SYMBOL
	|	'(' const_expr ')'
	;

