//
//	The BSL Compiler Kit
//	A Portable BSL-Compiler
//	Copyright (C) 1995 Nils M. Holm
//
//
//	THE DECLARATION PARSER
//
//	Parses declaration statements, and builds the symbol table.
//
//	Public functions:
//		SY_INIT()	Initialize the symbol table routines
//		FINDSYM()	Find a symbol in the symbol table
//		ADDGLOB()	Create a global symbol table entry
//		ADDLOCL()	Create a local symbol table entry
//		CLRLOCALS()	Clear local symbols
//		DECLARE()	Parse a declaration
//		D_INIT()	Allocate and initialize one word
//		FNDECL()	Parse a function declaration
//		FWDDECL()	Parse a FORWARD declaration
//		STCDECL()	Parse a STRUCTURE declaration
//		STCSIZE()	Determine the size of a structure
//

#include( blib.hdr)

#include( defs.hdr)
#include( extrn.hdr)
#include( error.hdr)
#include( token.hdr)
#include( glue.hdr)
#include( string.hdr)

var	hashtab . HASHSIZE;

extern	report, glbsave, locsave,
	newnode, expr, cexpr,
	mkstring, mktable, link, cut,
	prepend, next, xsymname,
	match, xlparen, xrparen,
	xlbrace, xrbrace, xsemi,
	function, newlabel, freeloc,
	unreferenced;


//	SY_INIT()
//	Purpose:	Initialize the symbol table
//	Receives:	NNOD	- The number of nodes allocated. Used to
//				  compute the remaining free space.
//	Returns:	-
//
public sy_init(nnod) {
	var	nsiz = nnod*NODE, i;

	Symbols = Memory + nsiz;
	Stsize = (MEMTOP - nsiz) / SYM * SYM;
	St0 = 0;
	St1 = Stsize-SYM;
	Forwp = 0;
	for (i=0, HASHSIZE) hashtab.i = 0;
}


//	NEWGLOB()
//	Purpose:	Allocates an entry in the global symbol table.
//	Receives:	-
//	Returns:	A pointer to a free entry
//
newglob() {
	if (St0 U>= St1) report(ESYMENT, FATAL);

	return(Symbols + (St0 += SYM) - SYM);
}


//	NEWLOCL()
//	Purpose:	Allocates an entry in the local symbol table.
//	Receives:	-
//	Returns:	A pointer to a free entry
//
newlocl() {
	if (St0 U>= St1) report(ESYMENT, FATAL);

	return(Symbols + (St1 -= SYM) + SYM);
}


//	FINDSYM(NAME, GLOB)
//	Purpose:	Finds a symbol in the symbol table.
//	Receives:	NAME - Name of symbol to search
//			GLOB - Search global symbol table
//	Returns:	A pointer to the entry found, or NULL,
//			when the symbol does not exist.
//
public findsym(name, glob) {
	var	si;
	var	s, e;
	var	p = 0, l, k = 0;

	s = (glob-> 0, St1+SYM) + Symbols;
	e = (glob-> St0, Stsize) + Symbols;
	l = glob-> 0, Level<1-> 1, Level;

	while (name.k) {
		p += name.k*k; k += 1;
	}
	p %= HASHSIZE;
	if (hashtab.p && hashtab.p.S_level == l && hashtab.p.S_name &&
	strequ(hashtab.p.S_name, name))
		return hashtab.p;

	ufor (si=s, e, SYM) {
		if (.(si.S_name) == .name && strequ(si.S_name, name))
			return hashtab.p = si;
	}

	return 0;
}


//	ADDGLOB(NAME, FLAGS, SIZE, STDECL)
//	Purpose:	Adds a global symbol to the symbol table.
//			When STDECL is FALSE, ADDGLOB() checks for
//			duplicate globals.
//	Receives:	NAME	- the symbol name
//			FLAGS	- the symbol flags (see DEFS.H)
//			SIZE	- the size of the associated object.
//			STDECL	- the new symbol is part of a
//				structure definition.
//	Returns:	A pointer to the new entry.
//
public addglob(name, flags, size) {
	var	new;

	if (findsym(name, 1)) {
		report(EDGLOB, ERROR); return 0;
	}

	if (.name == 'S' && strequ(name, "START")) flags |= (PUB|REFD);

	new = newglob();
	new.S_name = glbsave(name);
	new.S_addr = Dtop;
	new.S_size = size;
	new.S_flags = flags;
	new.S_level = 0;
	return new;
}


//	ADDLOCL(NAME, FLAGS, SIZE)
//	Purpose:	Adds a new symbol to the local symbol table.
//			ADDLOCL() check for duplicate locals on the
//			same level, and shadowing of global symbols.
//	Receives:	NAME	- the symbol name
//			FLAGS	- the symbol flags (see DEFS.H)
//			SIZE	- the size of the associated object.
//	Returns:	A pointer to the new entry.
//
public addlocl(name, flags, size) {
	var	new, old;
	var	levl = Level<1-> 1, Level;

	if (old = findsym(name, 0)) {
		if (old.S_level == levl) {
			report(EDLOCL, ERROR); return 0;
		}
	}
	else if (findsym(name, 1)) {
		report(ESHAD, WARNING);
	}

	new = newlocl();
	new.S_name = locsave(name);
	new.S_addr = (flags & STAT)-> Dtop, Ftop;
	new.S_size = size;
	new.S_flags = flags;
	new.S_level = levl;
	return new;
}


//	CLRLOCALS(N)
//	Purpose:	Clear local symbols of all levels down to N.
//	Receives:	N	- The level to clear
//	Returns:	yes.
//
public clrlocals(n) {
	while (St1 U< Stsize-SYM && (Symbols + St1+SYM).S_level >= n) {
		St1 += SYM;
		if (!((Symbols+St1).S_flags & REFD))
			unreferenced(Symbols+St1, "local");
		(Symbols+St1).S_name = 0;
	}
	if (!n) freeloc();
}


//	VECSIZE()
//	Purpose:	Parse the `size' part of a vector declaration.
//	Receives:	-
//	Returns:	the size of the vector.
//
vecsize() {
	var	size;

	if ((size = cexpr()) < 1) {
		report(ESIZE, ERROR);
		size = 1;
	}
	return size;
}


//	ALLOC(SIZE, GLOBL)
//	Purpose:	Allocates static memory/stack space.
//	Receives:	SIZE	- number of words to allocate
//			GLOBL	- allocate global/local space
//	Returns:	-
//
public alloc(size, globl) {
	if (globl) {
		if ((Dtop += size) U>= DSMAX) report(EDSOVF, FATAL);
	}
	else {
		if ((Ftop += size) U>= STKLEN) report(ESTKOVF, WARNING);
	}
}


//	D_INIT(T, V)
//	Purpose:	Allocates one word, and emit code to initialize it.
//	Receives:	TYPE	- type (CODE, DATA)
//			V	- initial value
//	Returns:	-
//
public d_init(type, v) {
	var	t;

	t = newnode(G_INI);
	t.N_val = 1;

	t = prepend(t, G_NUM);
	t.N_val = Dtop;

	if (type != NONE) {
		t = prepend(t, (type==CODE)-> G_RLA, G_REL);
	}

	t = prepend(t, G_NUM);
	t.N_val = v;

	alloc(1, 1);
	return t;
}


//	GLOB_INIT(SYM)
//	Purpose:	Parses the initialization of a global symbol.
//	Receives:	SYM	- A pointer to the symbol table entry.
//	Returns:	-
//
glob_init(sp) {
	var	t, addr;

	if (Token == LBRACK) {
		next();
		t = mktable(@addr);
		t = link(t, d_init(DATA, addr));
		sp.S_addr = Dtop-1;
	}
	else if (Token == STRING) {
		t = d_init(DATA, Dtop+1);
		t = link(t, mkstring());
	}
	else {
		t = d_init(NONE, cexpr());
	}
	return t;
}


//	LOCL_INIT(STATC)
//	Purpose:	Parses the initialization of a local symbol.
//	Receives:	STATC	- the symbol is STATIC
//	Returns:	A parse tree containing the required
//			instructions.
//
locl_init(statc) {
	var	root;

	if (statc) {
		root = newnode(G_SG);
		root.N_val = Dtop;
		root.N_right = newnode(G_DRO);
	}
	else {
		root = newnode(G_PSH);
	}
	alloc(1, statc);
	root.N_left = expr(0);
	return root;
}


//	GLOB_ALLOC(SIZE, VEC)
//	Purpose:	Allocates space for a global variable, or a vector.
//	Receives:	SIZE	- vector size
//			VEC	- the space is for a vector
//	Returns:	-
//
glob_alloc(size, vec) {
	var	t;

	t = vec-> d_init(DATA, Dtop+1), 0;
	alloc(size, 1);
	return t;
}


//	LOCL_ALLOC(SIZE, VEC, STATC)
//	Purpose:	Allocates space for a local variable, or a vector.
//	Receives:	SIZE	- vector size
//			VEC	- the space is for a vector
//			STATC	- the associated symbol is STATIC
//	Returns:	A parse tree that contains the instructions for
//			the initialization.
//
locl_alloc(size, vec, statc) {
	var	root, t;

	if (statc) return(glob_alloc(size, vec));

	if (!vec) {
		root = newnode(G_PSH);
		t = newnode(G_LN);
		t.N_val = 0;
		root.N_left = t;
		alloc(1, 0);
	}
	else {
		root = newnode(G_VEC);
		root.N_val = size;
		root.N_left = newnode(G_ADJ);
		root.N_right = newnode(G_PSH);

		t = root.N_left;
		t.N_left = newnode(G_LN);
		t = t.N_left;
		t.N_val = Ftop+size;

		alloc(size+1, 0);
	}

	return root;
}


//	DECLARE(GLOBL, FLAGS)
//	Purpose:	Parse the symbol list of a declaration statement,
//			and adds symbols to the symbol table.
//	Receives:	GLOBL	- declaration is in global context
//			FLAGS	- flags for the new symbols
//	Returns:	global context:	NULL
//			local context:	A parse tree that contains
//			the instructions needed for the initialization.
//
public declare(globl, flags) {
	var	root = 0, t;
	var	new;
	var	sname.NAMELEN+1;
	var	size, vec;

	if (globl) flags |= STAT;

	while (1) {
		vec = 0;
		bcopy(Strval, sname, NAMELEN);
		sname.NAMELEN = 0;
		xsymname();

		if (Token == LV && !(flags & RVAL)) {
			next();
			size = vecsize();
			vec = 1;
		}
		else {
			size = 1;
			vec = 0;
		}

		if (globl)
			new = addglob(sname, flags, size);
		else
			new = addlocl(sname, flags, size);

		if (flags & RVAL) {
			match(MOV, "'='", ERROR);
			new.S_addr = cexpr();
		}
		else if (Token == MOV) {
			if (size != 1 || flags & EXT) report(EINIT, ERROR);
			next();
			if (globl)
				t = glob_init(new);
			else
				t = locl_init((flags & STAT) != 0);
			root = link(root, t);
		}
		else {
			if (!(flags & EXT)) {
				if (globl)
					t = glob_alloc(size, vec);
				else
					t = locl_alloc(size, vec,
					(flags & STAT) != 0);
				root = link(root, t);
			}
		}

		if (Token != COMMA) break;
		next();
	}

	root = cut(root);
	xsemi();

	return root;
}


//	PARAMETERS()
//	Purpose:	Parse the parameter list of a function declaration,
//			and add the parameter names to the symbol table.
//	Receives:	-
//	Returns:	The number of parameters
//
parameters() {
	var	locals = St1;
	var	i, j, nparm = 0;
	var	sname.NAMELEN+1;

	xlparen();
	if (Token != RPAREN) while (1) {
		bcopy(Strval, sname, NAMELEN);
		sname.NAMELEN = 0;
		xsymname();
		addlocl(sname, 0, 1);
		if (Token != COMMA) break;
		next();
		nparm += 1;
	}
	xrparen();
	j = nparm+1;
	ufor (i=St1+SYM, locals+SYM, SYM) {
		(Symbols+i).S_addr = -j-1;
		j -= 1;
	}
	return nparm;
}


//	FNDECL(FLAGS)
//	Purpose:	Parses a function declaration, and add the
//			function name to the symbol table.
//	Receives:	FLAGS	- The flags for the function symbol.
//	Returns:	-
//
public fndecl(flags) {
	var	name.NAMELEN+1;
	var	label, forwrd = 0;
	var	fn;

	bcopy(Strval, name, NAMELEN);
	name.NAMELEN = 0;
	xsymname();
	if (fn = findsym(name, 1)) {
		if (fn.S_flags & FORW) {
			label = fn.S_size;
			fn.S_flags &= ~FORW;
			fn.S_flags |= (STAT|flags);
		}
		else {
			report(EDGLOB, ERROR);
		}
		forwrd = 1;
	}
	else {
		fn = addglob(name, STAT|flags, 0);
		label = newlabel();
	}

	parameters();
	function(label, forwrd);
}


//	FWDDECL()
//	Purpose:	Parses the symbol list of a FORWARD declaration,
//			and adds the symbols to the symbol table.
//	Receives:	-
//	Returns:	-
//
public fwddecl() {
	var	sname.NAMELEN+1;
	var	sp;
	var	root = 0, t;

	while (1) {
		bcopy(Strval, sname, NAMELEN);
		sname.NAMELEN = 0;
		xsymname();
		sp = addglob(sname, FORW|STAT, 0);
		if (Forwp >= MAXFORW) report(ENFORW, FATAL);
		t = d_init(CODE, sp.S_size = newlabel());
		Forwtab.Forwp = Sp.S_size; Forwp += 1;
		if (!root)
			root = t;
		else
			root = link(root, t);
		if (Token != COMMA) break;
		next();
	}
	xsemi();
	return root;
}


//	STCDECL()
//	Purpose:	Parses the body of a STRUCTURE declaration,
//			and adds the struture with all its members to
//			the symbol table.
//	Receives:	-
//	Returns:	-
//
public stcdecl() {
	var	sname.NAMELEN+1;
	var	base, mem;
	var	size, offset = 0;

	bcopy(Strval, sname, NAMELEN);
	sname.NAMELEN = 0;
	xsymname();
	if (findsym(sname, 1)) report(EDGLOB, ERROR);
	base = addglob(sname, STCT|RVAL, 0);
	xlbrace();

	while (1) {
		bcopy(Strval, sname, NAMELEN);
		sname.NAMELEN = 0;
		xsymname();
		if (findsym(sname, 1)) report(EDGLOB, ERROR);
		if (Token == LV) {
			next();
			size = vecsize();
		}
		else {
			size = 1;
		}
		mem = addglob(sname, RVAL, size);
		mem.S_addr = offset;
		offset += size;
		base.S_size += 1;
		if (Token != COMMA) break;
		next();
	}

	xrbrace();
	xsemi();
}


//	STCSIZE(STC)
//	Purpose:	Computes the size of a structure.
//	Receives:	A pointer to the symbol table entry of the
//			structure.
//	Returns:	the size.
//
public stcsize(stc) {
	var	i, j, size = 0;

	j = (stc.S_size+1)*SYM;
	ufor (i=SYM, j, SYM) size += (stc+i).S_size;
	return size;
}
