//
//	The BSL Compiler Kit
//	A Portable BSL-Compiler
//	Copyright (C) 1995 Nils M. Holm
//
//
//	ERROR HANDLER
//
//	All ERROR() calls go here.
//
//	Public functions:
//		REPORT()	report an error, and recover
//		EXPECT()	report a syntax error
//		UNDECLARED()	report an undeclared symbol
//		INTERNAL()	report a compiler error
//

#include( blib.hdr)

#include( defs.hdr)
#include( extrn.hdr)
#include( token.hdr)
#include( error.hdr)
#include( string.hdr)


var	_msg;

extern	next, addlocl, addglob;


//	RECOVER()
//	Purpose:	Resync the parser after a SEVERE error.
//			This routine could be greatly improved by
//			scanning for a resynch token, like
//			{ } , ; ( ) etc
//	Receives:	-
//	Returns:	-
//
recover() next();


//	REPORT(ID, CLASS)
//	Purpose:	Report/handle an error
//	Receives:	ID	- Error ID (see ERROR.H)
//			CLASS	- Error class
//	Returns:	Nothing. In case of a FATAL error,
//			it does not return at all.
//
public report(id, class) {
	var	msg, msgbuf.256/_BPW_, idbuf;

	switch (id) {
		case ESHAD:	msg = "local decl shadows global"; break;
		case ESTKOVF:	msg = "possible stack overflow"; break;
		case EOVFL:	msg = "numeric literal overflow"; break;
		case ETERM:	msg = "unterminated character literal"; break;
		case ELVAL:	msg = "bad lvalue"; break;
		case EUNREFD:
		case EUDECL:
		case ESYN:	msg = _msg; break;
		case EDECL:	msg = "invalid declaration"; break;
		case EDIV:	msg = "divide by zero"; break;
		case EDGLOB:	msg = "duplicate global"; break;
		case EDLOCL:	msg = "duplicate local"; break;
		case ESIZE:	msg = "bad size"; break;
		case EINIT:	msg = "invalid initialization"; break;
		case EBREAK:	msg = "'loop' or 'break' in wrong context";
				break;
		case EDCASE:	msg = "duplicate case"; break;
		case EXADDR:	msg = "address of external is unknown"; break;
		case EAADDR:	msg = "local address is not constant"; break;
		case EMEM:	msg = "insufficient memory"; break;
		case EIO:	msg = "I/O error"; break;
		case EINPUT:	msg = "cannot open input file"; break;
		case EOUTPUT:	msg = "cannot create output file"; break;
		case ESCAN:	msg = "bad input character"; break;
		case ETOKL:	msg = "token too long"; break;
		case ENODES:	msg = "out of free nodes"; break;
		case ENYI:	msg = "not yet implemented"; break;
		case ELABEL:	msg = "too many labels"; break;
		case ESYMENT:	msg = "symbol table overflow"; break;
		case EDSOVF:	msg = "data segment overflow"; break;
		case ETBLL:	msg = "table too big"; break;
		case ENCASE:	msg = "too many cases"; break;
		case ENFORW:	msg = "too many forward declarations"; break;
		case ECOMP:	msg = "compiler error"; break;
		default:	msg = "unknown error"; class = FATAL; break;
	}

	strcpy(msgbuf, Inname);
	strcat(msgbuf, " (");
	strcat(msgbuf, itoa(Lno, 0));
	strcat(msgbuf, "): ");
	strcat(msgbuf, msg);
	idbuf = " [x"; idbuf.2 = class;
	strcat(msgbuf, idbuf);
	strcat(msgbuf, itoa(id, 0));
	strcat(msgbuf, "]");
	strcat(msgbuf, _NLCHAR_);
	writes(SYSERR, MSGBUF);

	if (class == WARNING) return;

	Errcount += 1;

	if (class == SEVERE) recover();

	if (class == FATAL) {
		writes(SYSERR, "Terminating.");
		writes(SYSERR, _NLCHAR_);
		halt 2;
	}
}


//	EXPECT(WHAT, CLASS)
//	Purpose:	Report a syntax error
//	Receives:	WHAT	- A description of the expected token
//			CLASS	- The class of the error
//	Returns:	thru REPORT()
//
public expect(what, class) {
	var	buf.TOKLEN*2/_BPW_;

	strcpy(buf, "syntax error (");
	strcat(buf, what);
	strcat(buf, " expected)");
	_msg = buf;
	report(ESYN, class);
}


//	UNDECLARED(NAME)
//	Purpose:	Report reference to an undeclared symbol
//	Receives:	NAME	- Name of undeclared symbol
//	Returns:	thru REPORT()
//
public undeclared(name) {
	var	buf.256/_BPW_;
	var	fix;

	fix = (Level-> addlocl, addglob)(name, Level-> 0, STAT, 1);
	strcpy(buf, "undeclared symbol (");
	strcat(buf, name);
	strcat(buf, ")");
	_msg = buf;
	report(EUDECL, ERROR);
	return fix;
}


//	UNREFERENCED(SP, SCOPE)
//	Purpose:	Report an unused global/local symbol
//	Receives:	NAME	- Name of undeclared symbol
//	Returns:	thru REPORT()
//
public unreferenced(sp, scope) {
	var	buf.256/_BPW_;

	strcpy(buf, "unreferenced ");
	strcat(buf, scope);
	strcat(buf, " symbol: ");
	strcat(buf, sp.S_name);
	_msg = buf;
	report(EUNREFD, WARNING);
}


//	INTERNAL(ID)
//	Purpose:	Report a compiler error (FATAL)
//	Receives:	ID	- A unique ID
//	Returns:	Never, because it causes a FATAL error
//
public internal(id) {
	writes(SYSERR, "Internal error #");
	writes(SYSERR, itoa(id, 0));
	writes(SYSERR, _NLCHAR_);
	report(ECOMP, FATAL);
}
