//
//	The BSL Compiler Kit
//	A Portable BSL-Compiler
//	Copyright (C) 1995 Nils M. Holm
//
//
//	UTILITIES
//
//	Miscellaneous functions.
//
//	Public functions:
//		NEWNODE()	allocate a node
//		FREENODE()	release an allocated node
//		CLEARNODES()	release all nodes
//		GLBSAVE()	save a global variable name
//		LOCSAVE()	save a local variable name
//		FREELOC()	remove all local variable names
//		NEWLABEL()	generate a unique label
//		MATCH()		match a token
//		XLPAREN()	match '('
//		XRPAREN()	match ')'
//		XLBRACE()	match '{'
//		XRBRACE()	match '}'
//		XSEMI()		match ';'
//		XSYMNAME()	match a symbol name
//		XCOMMA()	match ','
//		XCOLON()	match ':'
//		LINK()		link two trees
//		PREPEND()	append an object to a list
//		CUT()		remove superflous glue nodes
//		FINDSX()	find the table index of a symbol
//		FETCH()		generate indirection code
//		EMIT_SYM()	emit symbol names
//		UT_INIT()	initialize the utility routines
//

#include( blib.hdr)
#include( defs.hdr)
#include( extrn.hdr)
#include( error.hdr)
#include( token.hdr)
#include( glue.hdr)
#include( string.hdr)


extern	next, expect, report,
	n_emit, t_emit, do_emit;


var	Label;				// Next Label to use
var	Freelist;			// Free list of node pool
var	Usedlist,			// List of used nodes
	Lastused;			// Last used node
var	Gsptr, Lsptr;			// Global/Local string pool pointer


//	NEWNODE()
//	Purpose:	Allocate a node, perform garbage collection,
//			if neccessary.
//	Receives:	TOKEN	- The desired type.
//	Returns:	A pointer to the allocated node
//
public newnode(tok) {
	var	new;

	if (Freelist == 0) {
		var	i, np;

		ufor (i=0, Nsize, NODE) {
			np = Nodes+i;
			if (np.N_token == UNUSED) {
				np.N_token = FREE;
				np.N_link = Freelist;
				Freelist = np;
			}
		}
		Usedlist = Lastused = 0;
		ufor (i=0, Nsize, NODE) {
			np = Nodes+i;
			if (np.N_token != FREE) {
				np.N_link = Usedlist;
				Usedlist = np;
				if (!Lastused) Lastused = np;
			}
		}
		if (Freelist == 0) report(ENODES, FATAL);
	}

	new = Freelist;
	Freelist = Freelist.N_link;
	new.N_token = tok;
	new.N_left = new.N_right = 0;
	new.N_link = Usedlist;
	Usedlist = new;
	if (!Lastused) Lastused = new;
	return new;
}


//	FREENODE()
//	Purpose:	Release an allocated node.
//	Receives:	NODE	- A pointer to the node.
//	Returns:	-
//
public freenode(np) np.N_token = UNUSED;


//	CLEARNODES()
//	Purpose:	Release all used nodes to the node pool
//	Receives:	-
//	Returns:	-
//
public clearnodes() {
	var	np;

	if (!Lastused) return;
	np = Usedlist;
	while (np) {
		np.N_token = FREE;
		np = np.N_link;
	}
	Lastused.N_link = Freelist;
	Freelist = Usedlist;
	Lastused = Usedlist = 0;
}


//	INITNODES()
//	Purpose:	Initialize the node pool
//	Receives:	-
//	Returns:	-
//
public initnodes() {
	var	i, np;

	ufor (i=0, Nsize, NODE) {
		np = Nodes+i;
		np.N_link = np+NODE;
		np.N_token = FREE;
	}
	np.N_link = 0;
	Freelist = Nodes;
	Lastused = Usedlist = 0;
}


//	GLBSAVE(S)
//	Purpose:	Allocate string space, and copy the given string
//			to the allocated space.
//	Receives:	S	- The string to save
//	Returns:	A pointer to the copy S.
//
public glbsave(s) {
	var	new, l;

	while (s.l) l+=1;
	l += 1;
	if (Gsptr + l U>= Lsptr) report(EMEM, FATAL);
	new = Gsptr;
	strcpy(Gsptr, s);
	Gsptr += l;
	return new;
}


//	LOCSAVE(S)
//	Purpose:	Allocate string space, and copy the given string
//			to the allocated space.
//	Receives:	S	- The string to save
//	Returns:	A pointer to the copy S.
//
public locsave(s) {
	var	l;

	while (s.l) l+=1;
	l += 1;
	Lsptr -= l;
	if (Lsptr U<= Gsptr) report(EMEM, FATAL);
	strcpy(Lsptr, s);
	return Lsptr;
}


//	FREELOC()
//	Purpose:	Free the string space used for local variables.
//	Receives:	-
//	Returns:	-
//
public freeloc() Lsptr = Str_pool + STRSPACE-1;


//	NEWLABEL()
//	Purpose:	Generate a unique label.
//	Receives:	-
//	Returns:	A new label.
//
public newlabel() {
	if (Label > MAXLBL) report(ELABEL, FATAL);
	return (Label+=1)-1;
}


//	MATCH(TOK, WHAT, CLASS)
//	Purpose:	Match a token, and generate a syntax error
//			on failure.
//	Receives:	TOK	- required token.
//			WHAT	- verbose description of TOK
//			CLASS	- class of error when the match
//				fails.
//	Returns:	-
//
public match(tok, what, class) {
	if (tok == Token)
		next();
	else
		expect(what, class);
}


//	XLPAREN()
//	Purpose:	Match a left paren [(].
//	Receives:	-
//	Returns:	-
//
public xlparen() match(LPAREN, "'('", ERROR);


//	XRPAREN()
//	Purpose:	Match a right paren [)].
//	Receives:	-
//	Returns:	-
//
public xrparen() match(RPAREN, "')'", ERROR);


//	XLBRACE()
//	Purpose:	Match a left brace ({).
//	Receives:	-
//	Returns:	-
//
public xlbrace() match(LBRACE, "'{'", ERROR);


//	XRBRACE()
//	Purpose:	Match a right brace (}).
//	Receives:	-
//	Returns:	-
//
public xrbrace() match(RBRACE, "'}'", ERROR);


//	XSEMI()
//	Purpose:	Match a semicolon (;).
//	Receives:	-
//	Returns:	-
//
public xsemi() match(SEMI, "';'", ERROR);


//	XSYMNAME()
//	Purpose:	Match a symbol name.
//	Receives:	-
//	Returns:	-
//
public xsymname() match(SYMBOL, "symbol name", ERROR);


//	XCOMMA()
//	Purpose:	Match a comma (,).
//	Receives:	-
//	Returns:	-
//
public xcomma() match(COMMA, "','", ERROR);


//	XCOLON()
//	Purpose:	Match a colon (:).
//	Receives:	-
//	Returns:	-
//
public xcolon() match(COLON, "':'", ERROR);


//	LINK(ROOT, NEW)
//	Purpose:	Link two nodes using a GLUE node.
//	Receives:	ROOT	- the root node of the parse tree
//			NEW	- the node to add
//	Returns:	A new parse tree, where NEW is pasted to
//			the head of ROOT (thru a GLUE node).
//
public link(root, new) {
	var	t;

	t = newnode(GLUE);
	t.N_left = root;
	t.N_right = new;
	return t;
}


//	PREPEND(NP, TYPE)
//	Purpose:	Create a new node and append it to the head of
//			a node list.
//	Receives:	NP	- the head of the list
//			TYPE	- the type (token) of the new head
//	Returns:	A new list with a type TYPE node as head.
//
public prepend(np, type) {
	var	t;

	t = newnode(type);
	t.N_left = np;
	return t;
}


//	CUT(ROOT)
//	Purpose:	Removes a superflous GLUE node from the
//			head of a parse tree (if any).
//	Receives:	ROOT	- the parse tree to cut
//	Returns:	The (possibly cut) parse tree.
//
public cut(root) {
	var	t;

	if (root.N_token == GLUE && root.N_left == 0) {
		t = root;
		root = root.N_right;
		freenode(t);
	}
	return root;
}


//	FINDSX(NAME)
//	Purpose:	Find the symbol table index of a given symbol.
//	Receives:	NAME	- the symbol to find.
//	Returns:	The index of the symbol,
//			or -1 if the symbol does not exist.
//
public findsx(name) {
	var	i, sp;

	ufor (i=0, St0, SYM) {
		sp = Symbols+i;
		if (sp.S_flags & STCT)
			i += sp.S_size*SYM;
		else if (.name == .(sp.S_name) &&
		strequ(name, sp.S_name))
			return i/SYM;
	}
	return -1;
}


//	FETCH(NP)
//	Purpose:	Convert a temporary SYM node into a load operation.
//			locals	-->	LL
//			globals	-->	LG
//			externs	-->	LX
//	Receives:	NP	- The SYM node
//	Returns:	Nothing, but modifies NP directly.
//
public fetch(np) {
	var	sp;

	if (np.N_token == G_SYM) {
		sp = np.N_val;
		if (sp.S_flags & EXT) {
			np.N_token = G_LX;
			np.N_val = findsx(sp.S_name);
		}
		else {
			np.N_token = (np.N_flags & GLOBAL ||
			sp.S_flags & STAT)-> G_LG, G_LL;
			np.N_val = sp.S_addr;
		}
	}
}


//	EMIT_SYM(FLAGS)
//	Purpose:	Create a PUB/EXT symbol table (BCODE).
//	Receives:	FLAGS	- desired flags
//			When FLAGS & PUBLIC, build a PUB table.
//			When FLAGS & EXTERN, build an EXT table.
//	Returns:	-
//
public emit_sym(flags) {
	var	i, nsym = 0;

	t_emit((flags==PUB)-> G_PUB, G_EXT);
	if (flags == EXT) flags |= REFD;

	for (i=0, St0, SYM) {
		if ((Symbols+i).S_flags & STCT)
			i += (Symbols+i).S_size * SYM;
		else if (((Symbols+i).S_flags & flags) == flags)
			nsym += 1;
	}
	n_emit(nsym);
	for (i=0, St0, SYM) {
		if ((Symbols+i).S_flags & STCT)
			i += (Symbols+i).S_size * SYM;
		else if (((Symbols+i).S_flags & flags) == flags) {
			do_emit((Symbols+i).S_name);
			n_emit((flags & EXT)-> i/SYM, (Symbols+i).S_addr);
		}
	}
}


//	UT_INIT()
//	Purpose:	Initialize the utility routines.
//	Receives:	NNOD	- The number of nodes to allocate
//	Returns:	-
//
public ut_init(nnod) {
	Label = 0;
	Nodes = Memory;
	Nsize = nnod * NODE;
	initnodes();
	Gsptr = Str_pool;
	freeloc();
}
