///////////////////////////////////////////////////////////////////////////////
//
//	BCR -- Resolve external links in BCODE objects
//	Copyright (C) 1995 Nils M. Holm
//
//	BCR attempts to resolve external links in BCODE modules.
//	When a link cannot be resolved within the input module,
//	it searches the library index created by NDXGEN for the
//	requested name, and loads the module containing the name
//	from the BCODE library.
//	BCR reads the input module from the standard input stream
//	(STDIN), and writes a loadable BCODE module to the standard
//	output (STDOUT). The output of BCR can directly be passed
//	to CG, the code generator.
//	When a link cannot be resolved, neither internally nor in
//	the library, BCR prints an error message. All messages
//	will be sent to the standard error stream (STDERR).
//
//	USAGE: BCR [-LRSV] [<DIRECTORY>]
//
//
//	-L		Just list the names of required modules.
//			Do not actually load any modules.
//			(with -V: list full path names)
//
//	-R		Reorder modules. Place library objects before
//			user objects in the output stream. This is a
//			little slower, but may help when the code
//			generator runs out of space, because it reduces
//			the number of forward links in the BCODE file.
//
//	-S		Do not load the startup module. Used to create
//			standalone executables.
//
//	-V		Report names of required modules to STDERR.
//
//	<DIRECTORY>	is the directory containing the library
//			index, and possibly the library itself,
//			too.
//

#include( blib.hdr)
#include( ../ios/ios.hdr)
#include( ../mem/mem.hdr)
#include( ../str/str.hdr)
#include( ../sys/syscall.hdr)

#include( ../opt.hdr)


value	POOLSZ	= 24576;
value	RECMAX	= 256;

value	PUBREC	= 'P';
value	EXTREC	= 'X';
var	BSTART	= "b0.bc0";
var	TMPFILE	= "bcr00000";

var	O_verbose	= 0,
	O_justlist	= 0,
	O_reorder	= 0,
	O_nostart	= 0;

var	Lc = 0;


structure LIST {
	L_item,
	L_next
};

structure LIBNDX {
	N_module,
	N_pubsym,
	N_extref,
	N_next
};


var	Publist = 0,
	Extlist = 0;

var	Index = 0;

var	Libpath = "c:/v/bsl/lib";


quit(n) {
	stdclose();
	halt n;
}


badrecord(recn) {
	swritef(STDERR, "bcr: invalid index record (#%ud)\r\n", recn);
	quit(1);
}


nomem() {
	sputl(STDERR, "bcr: out of memory\r\n");
	quit(1);
}


strsave(s) {
	var	new;

	if (!(new = malloc(strlen(s)+1))) nomem();
	strcpy(new, s);
	return new;
}


newlist(listp, name) {
	var	new;

	if (!(new = malloc(LIST))) nomem();
	new.L_item = strsave(name);
	new.L_next = .listp;
	.listp = new;
}


list_lookup(lp, name) {
	while (lp) {
		if (!strcmp(lp.L_item, name)) return lp;
		lp = lp.L_next;
	}
	return 0;
}


newndx(module, nlist, type) {
	var	new, name;
	var	t;

	if (type == PUBREC) {
		if (!(new = malloc(LIBNDX))) nomem();
		new.N_module = strsave(module);
		new.N_pubsym = new.N_extref = 0;
	}
	else {
		if (strcmp(module, Index.N_module)) {
			swritef(STDERR, "bcr: bad index order: %s\r\n",
			module);
			quit(1);
		}
		new = Index;
	}

	name = nlist;
	while (.nlist != '\n') {
		if (#iswhite(.nlist)) {
			t = .nlist; .nlist = 0;
			if (type == PUBREC)
				newlist(new+N_pubsym, name);
			else
				newlist(new+N_extref, name);
			.nlist = t;
			name = nlist+1;
			if (.nlist != '\n') nlist += 1;
		}
		else {
			nlist += 1;
		}
	}
	if (.name != '\n') {
		.nlist = 0;
		if (type == PUBREC)
			newlist(new+N_pubsym, name);
		else
			newlist(new+N_extref, name);
	}

	if (type == PUBREC) {
		new.N_next = Index;
		Index = new;
	}
}


readindex() {
	var	lpath . _PATHL_;
	var	buf   . RECMAX;
	var	lx;
	var	nlist, recn = 0;

	strfmt(lpath, "%s/Index", Libpath);
	if ((lx = sopen(lpath, S_READ)) == ENDSTREAM) {
		swritef(STDERR, "bcr: no library index: %s\r\n", lpath);
		quit(1);
	}

	sgetl(lx, buf, RECMAX);
	while (!#seof(lx)) {
		recn += 1;
		if (!(nlist = strchr(buf, '\t'))) {
			badrecord(recn);
			quit(1);
		}
		.nlist = 0;
		nlist += 1;
		while (#iswhite(.nlist)) nlist += 1;
		if (.buf == PUBREC)
			newndx(buf+2, nlist, PUBREC);
		else if (.buf == EXTREC)
			newndx(buf+2, nlist, EXTREC);
		else
			badrecord(recn);
		sgetl(lx, buf, RECMAX);
	}

	sclose(lx);
}


readword(src) {
	value	WORDLEN = 20;
	static var	buf.WORDLEN;
	var	c, i = 0;

	if (#seof(src)) return ENDSTREAM;

	c = #sgetc(src);
	while (#iswhite(c)) c = #sgetc(src);
	while (!#iswhite(c) && !#seof(src)) {
		buf.i = c; i += 1;
		if (i >= WORDLEN) {
			sputl(STDERR, "bcr: token too long\r\n");
			quit(1);
		}
		c = #sgetc(src);
	}
	buf.i = 0;

	return buf;
}


emitword(dest, word) {
	if (O_justlist) return;
	if (Lc >= 72) {
		sputl(dest, _NLCHAR_); Lc = 0;
	}
	if (Lc) #sputc(dest, '\s');
	sputl(dest, word);
	Lc += strlen(word)+1;
}


addlist(dest, listp) {
	var	n;
	var	name;

	n = stob(name = readword(STDIN), 10);
	emitword(dest, name);
	while (n) {
		if ((name = readword(STDIN)) == ENDSTREAM) {
			sputl(STDERR, "bcr: unexpected EOF\r\n");
			quit(1);
		}
		emitword(dest, name);
		newlist(listp, name);
		emitword(dest, readword(STDIN));
		n -= 1;
	}
}


collect(dest) {
	var	word;

	while ((word = readword(STDIN)) != ENDSTREAM) {
		emitword(dest, word);
		if (!strcmp(word, "PUB"))
			addlist(dest, @Publist);
		else if (!strcmp(word, "EXT"))
			addlist(dest, @Extlist);
	}
}


copymodule(name) {
	var	mpath . _PATHL_;
	var	mod;
	var	ch;

	strfmt(mpath, "%s/%s", Libpath, name);
	if (O_justlist) {
		swritef(STDOUT, "%s\r\n", O_verbose-> mpath, name);
		ssync(STDOUT);
		return;
	}
	if (O_verbose) {
		swritef(STDERR, "+ %s\r\n", name);
		ssync(STDERR);
	}
	if ((mod = sopen(mpath, S_READ)) == ENDSTREAM) {
		swritef(STDERR, "bcr: no such module: %s\r\n", mpath);
		quit(1);
	}
	ch = #sgetc(mod);
	while (!#seof(mod)) {
		#sputc(STDOUT, ch);
		ch = #sgetc(mod);
	}
	sclose(mod);
}


loadmodule(mp) {
	var	lp;
	forward ndx_lookup;

	copymodule(mp.N_module);
	lp = mp.N_pubsym;
	while (lp) {
		newlist(@Publist, lp.L_item);
		lp = lp.L_next;
	}
	lp = mp.N_extref;
	while (lp) {
		ndx_lookup(lp.L_item);
		lp = lp.L_next;
	}
}


ndx_lookup(name) {
	var	ip = Index;
	var	pp;

	if (list_lookup(Publist, name)) return 1;

	while (ip) {
		pp = ip.N_pubsym;
		while (pp) {
			if (!strcmp(pp.L_item, name)) {
				loadmodule(ip);
				return 1;
			}
			pp = pp.L_next;
		}
		ip = ip.N_next;
	}
	return 0;
}


resolve() {
	var	ext;
	var	ucount = 0;

	ext = Extlist;
	while (ext) {
		if (!ndx_lookup(ext.L_item)) {
			swritef(STDERR, "bcr: can't resolve: %s\r\n",
			ext.L_item);
			ucount += 1;
		}
		ext = ext.L_next;
	}
	return ucount;
}


copyback(src) {
	var	c = #sgetc(src);

	while (!#seof(src)) {
		#sputc(STDOUT, c);
		c = #sgetc(src);
	}
}


usage() {
	sputl(STDERR, "usage: bcr [-lv] [directory]\r\n");
	quit(1);
}


start(argcount, argvec) {
	var	Pool . POOLSZ;
	var	argbuf  . 128;
	var	dest, err;

	mminit(Pool, POOLSZ);
	stdopen();

	while (argcount) {
		unpack(.argvec, argbuf);
		.argvec = argbuf;
		if (..argvec != '-') break;
		.argvec += 1;
		while (..argvec) {
			switch (..argvec) {
				case 'v': case 'V':
					O_verbose += 1; break;

				case 's': case 'S':
					O_nostart = 1; break;

				case 'r': case 'R':
					O_reorder = 1; break;

				case 'l': case 'L':
					O_justlist = 1; break;

				default:
					usage();
			}
			.argvec += 1;
		}
		argvec += 1;
		argcount -= 1;
	}

	if (argcount > 0) Libpath = .argvec;

	if (O_justlist) O_reorder = 0;

	if (O_reorder) {
		if ((dest = sopen(TMPFILE, S_WRITE|S_CREATE)) == ENDSTREAM) {
			sputl(STDERR, "bcr: cannot create workfile (-R)\r\n");
			quit(1);
		}
	}
	else {
		dest = STDOUT;
	}

	readindex();
	if (!O_nostart) copymodule(BSTART);
	collect(dest);
	err = resolve()-> 1, 0;

	if (O_reorder) {
		sclose(dest);
		if ((dest = sopen(TMPFILE, S_READ)) == ENDSTREAM) {
			sputl(STDERR, "bcr: cannot reopen workfile (-R)\r\n");
			quit(1);
		}
		copyback(dest);
		sclose(dest);
		sys_remove(TMPFILE);
	}

	quit(err);
}
