///////////////////////////////////////////////////////////////////////////////
//
//	BSLPP -- Preprocess BSL source files
//	Copyright (C) 1995 Nils M. Holm
//
//	Usage: BSLPP [-P] [-I DIR] [-DNAME[=VALUE]] FILE
//
//	-P	Pipe operation. Read input from STDIN, write
//		to STDOUT. A file name must still be given.
//
//	-I DIR	Search path for # include(). DIR will be searched
//		after the current path.
//
//	-D...	# define() NAME [with optional VALUE].
//
//	The BSL Preprocessor BSLPP resolves macros in BSL source programs.
//	A macro must be introduced with a hash character (#) in BSL. When
//	BSLPP encounters a #-sign, it attempts to expand the macro with
//	the name following in the input stream. If the following character
//	is not alphanumeric or an underscore, the hash sign will be passed
//	thru literally.
//	Macros may be one of the following types:
//
//		builtin, strings, parameterized.
//
//	This is a brief summary of the builtin macros:
//	(***	There should not be a space between '#' and the macro
//		name, but if it wouldn't be there, MPP might get
//		confused, because it doesn't know //-comments	***)
//
//	# DEFINE (name, val)
//		Define macro <name> with value <val>. Either string
//		macros or parameterized (function-style) macros may
//		be created using # define. Expands to nothing.
//		Examples: # define(foo, bar)  # define(add(a, b), ((a)+(b)))
//
//	# UNDEF (name)
//		Remove the definition of macro <name>. Expands to nothing.
//		Example: # undef(_version_)
//
//	# INCLUDE (file)
//		Expands to the contents of the <file>. The file will first
//		be searched in the current directory, and then in the
//		standard header path. When the file name begins with a
//		tilde (~), <file> will not be searched in the current
//		directory.
//		Example: # include(~blib.hdr)
//
//	# IFDEF (name)
//		When no macro <name> exists, the following lines will
//		be replaced with empty lines (processing state is changed
//		to FALSE). When <name> is defined, this macro does nothing
//		and the state remains unchanged. Since # IFDEF saves the
//		state of the previous context, it may be nested.
//		Example: # IFDEF (msdos)
//
//	# ELSE
//		Reverses the setting of the most recent # IFDEF. When
//		it was true (the previous text was included), it changes
//		to false (following text will not be included) and
//		vice versa. When the state of the surrounding context
//		is FALSE, # ELSE does nothing.
//
//	# ENDIF
//		Terminates a conditional block started with # IFDEF, and
//		restores the previous processing state.
//
//	# ERROR (message)
//		Prints an error message on the standard error device,
//		and terminates BSLPP. Used to generate user error messages.
//		Example: # error (This line should not be reached)
//
//	# COMMENT
//		Begins a comment. This is basically the same as an
//		# IFDEF(undefined_macro)
//		Comment blocks may be nested.
//
//	# CEND
//		Terminates a comment started with #COMMENT.
//
//
//	User-defined macros may be created with # define(). Macros
//	may be used to define symbolic names for constant value, or
//	to create small inline functions. Like builtins, user-defined
//	macros will only be expanded, when the leading '#' is present.
//
//	Examples:
//		# define (BUFSIZE, 1024)
//		# define (max(a, b), ((a) > (b)-> (a), (b)))
//
//		BUFSIZE		expands to	BUFSIZE
//		# BUFSIZE	expands to	1024
//		max(X, Y)	expands to	max(X, Y)
//		# max(X, Y)	expands to	((X) > ( Y)-> (X), ( Y))
//

#include( blib.hdr)
#include( ../ios/ios.hdr)
#include( ../mem/mem.hdr)
#include( ../str/str.hdr)

#include( ../opt.hdr)

value	POOLSIZE	= 16384;
value	CHUNKSIZE	= 128;
value	BUFSIZE		= 512;
value	MAXINC		= 5;
value	MAXARGS		= 10;
value	PMAGIC		= 0x7F;
value	STACKLEN	= 100;
value	FALSE		= 0,
	TRUE		= !FALSE;

var	Hdrpath = "c:/v/bsl/hdr";

forward	pp_include, pp_define, pp_undef,
	pp_ifdef, pp_ifndef, pp_else, pp_endif,
	pp_error, pp_comment, pp_endcomm;

structure MACRO {
	M_name,
	M_val,
	M_narg,
	M_flags,
	M_next
};

value	M_BUILTIN	= 00001,
	M_FUNCTION	= 00002;

var	Builtin = [
		[ "INCLUDE",	@pp_include,	M_BUILTIN	],
		[ "DEFINE",	@pp_define,	M_BUILTIN	],
		[ "UNDEF",	@pp_undef,	M_BUILTIN	],
		[ "IFDEF",	@pp_ifdef,	M_BUILTIN	],
		[ "IFNDEF",	@pp_ifndef,	M_BUILTIN	],
		[ "ELSE",	@pp_else,	M_BUILTIN	],
		[ "ENDIF",	@pp_endif,	M_BUILTIN	],
		[ "ERROR",	@pp_error,	M_BUILTIN	],
		[ "COMMENT",	@pp_comment,	M_BUILTIN	],
		[ "CEND",	@pp_endcomm,	M_BUILTIN	],
		[ "_VERSION_",	"100",		0		],
		[ "_BSL100_",	0,		0		],
		0
	];

var	O_pipe = 0;

var	Srcfile,
	Srcname,
	Srcline;

var	Outfile;
var	Lastch = ENDSTREAM;

var	Macros	= 0;
var	Inclevel = 0;

var	S_stack . STACKLEN, Sptr = 0, State = 1;
var	Comment = 0;

var	Buffer . BUFSIZE;


quit(n) {
	stdclose();
	halt n;
}


report(msg) {
	swritef(STDERR, "bslpp: %s (%ud): %s\r\n",
	Srcname, Srcline, msg);
	quit(1);
}


longch() report("token too long");


nomem() report("out of memory");


overflow() report("buffer overflow");


badeof() report("unexpected EOF");


nesting() report("#else/#endif with no matching #if");


expect(str) {
	var	msg . 128;

	strfmt(msg, "\"%s\" expected", str);
	report(msg);
}


undefined(name) {
	var	msg . 128;

	strfmt(msg, "undefined macro: \"%s\"", name);
	report(msg);
}


lcomment(src) {
	var	c = #sgetc(src);

	while (c != '\n' && c != '\r') {
		c = #sgetc(src);
	}
	return c;
}


readsrc(src, spp) {
	static var	chunk . CHUNKSIZE;
	var	c;
	var	i = 0;

	if (spp) .spp = 0;

	if (Lastch != ENDSTREAM) {
		c = Lastch; Lastch = ENDSTREAM;
	}
	else {
		c = #sgetc(src);
	}

	while (1) {
		if (#iswhite(c)) {
			if (spp) .spp =  '\s';
			if (!spp || c == '\r' || c == '\n')
				#sputc(Outfile, (c == '\t'-> '\s', c));
			while (#iswhite(c)) {
				if (c == '\r' || c == '\n') {
					#sputc(Outfile, c);
					if (c == '\n') Srcline += 1;
				}
				c = #sgetc(src);
			}
		}

		if (c == '/') {
			chunk.0 = c;
			if ((c = #sgetc(src)) == '/') {
				c = lcomment(src);
			}
			else {
				Lastch = c;
				chunk.1 = 0;
				return chunk;
			}
		}
		else {
			break;
		}
	}

	if (#seof(src)) return ENDSTREAM;

	if (#isalpha(c) || c == '_') {
		while (#isalpha(c) || #isdigit(c) || c == '_') {
			if (i >= CHUNKSIZE) longch();
			chunk.i = c; i += 1;
			c = #sgetc(src);
		}
		chunk.i = 0;
		Lastch = c;
	}

	else if (#isdigit(c)) {
		chunk.i = c; i += 1;
		c = #sgetc(src);
		if (c == 'x' || c == 'X') {
			chunk.i = c; i += 1;
			c = #sgetc(src);
		}
		while (#isdigit(c)) {
			if (i >= CHUNKSIZE) longch();
			chunk.i = c; i += 1;
			c = #sgetc(src);
		}
		chunk.i = 0;
		Lastch = c;
	}

	else if (c == '\'' || c == '"') {	// " <- (MPP quote bug)
		var	quote = c, escape = 1;

		while (c != quote || escape) {
			if (i >= CHUNKSIZE) longch();
			if (#seof(src) || c == '\n')
				report("unterminated character literal");
			chunk.i = c; i += 1;
			if (c == '\\' && !escape)
				escape = 1;
			else
				escape = 0;
			c = #sgetc(src);
		}
		chunk.i = quote; i += 1;
		chunk.i = 0;
	}

	else {
		chunk.0 = c;
		chunk.1 = 0;
	}

	return chunk;
}


readch(src) {
	var	c;

	if (Lastch != ENDSTREAM) {
		c = Lastch; Lastch = ENDSTREAM;
	}
	else {
		c = #sgetc(src);
	}
	if (c == '\r' || c == '\n') #sputc(Outfile, c);
	if (c == '\n') Srcline += 1;

	if (#seof(src)) return ENDSTREAM;

	return c;
}


skip() {
	var	c;

	c = readch(Srcfile);
	while (#iswhite(c)) {
		if (c == '\r' || c == '\n') #sputc(Outfile, c);
		if (c == '\n') Srcline += 1;
		c = #sgetc(Srcfile);
	}
	return c;
}


getstring(dlm) {
	var	c, i = 0;

	c = skip();
	while (c != dlm) {
		if (#seof(Srcfile)) badeof();
		if (c != '\r') {
			if (c == '\n') {
				#sputc(Outfile, c);
				c = '\s';
			}
			Buffer.i = c;
			if ((i += 1) >= BUFSIZE) overflow();
		}
		else {
			#sputc(Outfile, c);
		}
		c = #sgetc(Srcfile);
	}
	Buffer.i = 0;
	return Buffer;
}


getsymb() {
	var	c, i = 0;

	c = skip();
	while (#isalpha(c) || #isdigit(c) || (c) == '_') {
		Buffer.i = c;
		if ((i += 1) >= BUFSIZE) overflow();
		c = #sgetc(Srcfile);
	}
	Buffer.i = 0;
	Lastch = c;
	return Buffer;
}


strsave(s) {
	var	new;

	if (!(new = malloc(strlen(s)+1))) nomem();
	strcpy(new, s);
	return new;
}


upcase(p) {
	while (.p) {
		.p = #islower(.p)-> #toupper(.p), .p;
		p += 1;
	}
}


findmac(name) {
	var	mp = Macros;

	upcase(name);
	while (mp) {
		if (.name == .(mp.M_name) && !strcmp(name, mp.M_name))
			return mp;
		mp = mp.M_next;
	}
	return 0;
}


newmac(name, val) {
	var	new;

	if (!(new = malloc(MACRO))) nomem();
	upcase(name);
	new.M_name = strsave(name);
	if (val) new.M_val = strsave(val);
	new.M_flags = 0;
	new.M_next = Macros;
	Macros = new;
	return new;
}


initmac() {
	var	bp = Builtin;
	var	mac;

	while (.bp) {
		if ((.bp).2 == M_BUILTIN) {
			mac = newmac((.bp).0, 0);
			mac.M_val = .(.bp).1;
		}
		else {
			mac = newmac((.bp).0, (.bp).1);
		}
		mac.M_flags = (.bp).2;
		bp += 1;
	}
}


pp_include() {
	forward process;
	var	fullpath . _PATHL_;
	var	name, tilde = 0;
	var	oldfile,
		oldname,
		oldline;

	if (!State || Comment) return;

	if ((Inclevel += 1) >= MAXINC) report("too many nested includes");

	if (skip() != '(') expect("(");
	name = getstring(')');
	if (.name == '~') {
		tilde = 1; name += 1;
	}
	strfmt(fullpath, "%s/%s", Hdrpath, name);
	if (tilde) name = fullpath;

	oldfile = Srcfile;
	oldname = Srcname;
	oldline = Srcline;

	if ((Srcfile = sopen(name, S_READ)) == ENDSTREAM) {
		if ((Srcfile = sopen(fullpath, S_READ)) == ENDSTREAM) {
			var	noinc . 128;

			strfmt(noinc, "cannot open header: %s", name);
			report(noinc);
		}
	}
	else {
		strcpy(fullpath, name);
	}
	Srcname = fullpath;

	process();

	sclose(Srcfile);
	Srcfile = oldfile;
	Srcname = oldname;
	Srcline = oldline;
	swritef(Outfile, "# \"%s\" %ud%s", Srcname, Srcline, _NLCHAR_);

	Lastch = ENDSTREAM;
	Inclevel -= 1;
}


savearg(arg) {
	forward	expand;
	var	expb . BUFSIZE;

	while (#iswhite(arg)) arg += 1;
	if (.arg == '#')
		expand(arg+1, expb, 0);
	else
		expb = arg;
	return strsave(expb);
}


expand(source, dest, deltap) {
	var	srcbuf  . CHUNKSIZE;
	var	argbuf  . MAXARGS, na;
	var	mac, p, q, c, argp, i;
	var	level, ql, al, bl, esc;
	var	manyargs = "too many arguments";

	p = 0;
	while (#isalpha(source.p) || #isdigit(source.p) || source.p == '_')
		p += 1;
	c = source.p;
	source.p = 0;
	if (!(mac = findmac(source))) undefined(source);
	source.p = c;
	source += p;

	if (mac.M_flags & M_BUILTIN) {
		report("cannot expand builtin macro");
	}

	if (deltap) .deltap = p;

	if (!(mac.M_flags & M_FUNCTION)) {
		return mac.M_val;
	}

	if (!(.source)) {
		source = srcbuf;
		c = readch(Srcfile);
		p = 0;
		level = 0;
		esc = ql = al = 0;
		while (1) {
			if (#seof(Srcfile)) badeof();
			if (p >= CHUNKSIZE) overflow();
			source.p = c; p += 1;
			if (!(ql | al)) {
				if (c == '(') {
					level += 1;
				}
				else if (c == ')') {
					if (!(level -= 1)) break;
				}
			}
			if (!esc) {
				if (c == '"' && !al) ql = !ql;		// "
				if (c == '\'' && !ql) al = !al;
			}
			esc = (c == '\\' && !esc);
			c = readch(Srcfile);
		}
		source.p = 0;
		if (deltap) .deltap -= p;
	}

	na = p = 0;
	c = source.p; p += 1;
	argp = 0;
	esc = ql = al = bl = 0;
	while (1) {
		if (!(ql | al)) {
			if (c == '(') {
				level += 1;
			}
			else if (c == ')') {
				if (!(level -= 1)) break;
			}
			if (c == '[') {
				bl += 1;
			}
			else if (c == ']') {
				bl -= 1;
			}
			else {
				if (!#iswhite(c) && !argp) argp = source+1;
			}
		}
		if (!esc) {
			if (c == '"' && !al) ql = !ql;		// "
			if (c == '\'' && !ql) al = !al;
		}
		if (c == ',' && level < 2 && !(ql | al | bl)) {
			if (!argp) report("argument syntax error");
			if (na >= MAXARGS) report(manyargs);
			.(source+p-1) = 0;
			argbuf.na = savearg(argp); na += 1;
			.(source+p-1) = ',';
			argp = source+p;
		}
		esc = (c == '\\' && !esc);
		c = source.p; p += 1;
	}
	if (argp) {
		if (na >= MAXARGS) report(manyargs);
		.(source+p-1) = 0;
		argbuf.na = savearg(argp); na += 1;
		.(source+p-1) = ')';
	}

	if (deltap) .deltap += p;

	if (na != mac.M_narg)
		report("wrong number of arguments to macro call");

	p = q = 0;
	source = mac.M_val;
	while (source.p) {
		if (source.p == PMAGIC) {
			strcpy(dest+q, argbuf.(source.(p+=1)-'0'));
			q += strlen(dest+q);
			p += 1;
		}
		else
		{
			if (q >= BUFSIZE) overflow();
			dest.q = source.p;
			q += 1;
			p += 1;
		}
	}
	dest.q = 0;

	for (i=0, na) mfree(argbuf.i);

	return dest;
}


fnmacro(name) {
	var	pnames . MAXARGS;
	var	narg = 0, pp;
	var	ch, c, space;
	var	level = 1, ql, al, esc;
	var	val  . BUFSIZE, vp = 0,
		val2 . BUFSIZE, vp2;
	var	mp, dp;

	if ((c = skip()) != ')') {
		Lastch = c;
		ch = readsrc(Srcfile, 0);
		while (1) {
			if (!#isalpha(.ch) && .ch != '_')
				report("bad argument list");
			upcase(ch);
			if (narg >= MAXARGS)
				report("too many arguments");
			pnames.narg = strsave(ch);
			narg += 1;
			if ((c = skip()) != ',') break;
			ch = readsrc(Srcfile, 0);
		}
		if (c != ')') expect(")");
	}
	if (skip() != ',') expect(",");

	ch = readsrc(Srcfile, 0);
	while (1) {
		if (.ch == '(') {
			level += 1;
		}
		else if (.ch == ')') {
			if (!(level -= 1)) break;
		}
		if (space) {
			val.vp = '\s'; vp += 1;
		}

		if (#isalpha(.ch) || .ch == '_') {
			upcase(ch);
			for (pp=0, narg) {
				if (!strcmp(ch, pnames.pp)) break;
			}
			if (pp<narg) {
				val.vp = PMAGIC;
				val.(vp+1) = pp+'0';
				vp += 2;
			}
			else {
				if (vp + strlen(ch) >= BUFSIZE) overflow();
				strcpy(val+vp, ch);
				vp += strlen(val+vp);
			}
		}
		else {
			if (vp + strlen(ch) >= BUFSIZE) overflow();
			strcpy(val+vp, ch);
			vp += strlen(val+vp);
		}
		ch = readsrc(Srcfile, @space);
	}

	for (pp = 0, narg) mfree(pnames.pp);
	val.vp = 0;

	vp2 = vp = 0;

	esc = ql = al = 0;
	while (val.vp) {
		if (!esc) {
			if (val.vp == '"' && !al) ql = !ql;	// "
			if (val.vp == '\'' && !ql) al = !al;
		}
		esc = (val.vp == '\\' && !esc);
		if (val.vp == '#' && !(ql | al)) {
			var	expb . BUFSIZE;

			vp += 1;
			ch = expand(val+vp, expb, @dp);
			val += dp;
			strcpy(val2+vp2, ch);
			vp2 += strlen(val2+vp2);
		}
		else {
			val2.vp2 = val.vp;
			vp2 += 1;
			vp += 1;
		}
	}
	val2.vp2 = 0;

	mp = newmac(name, val2);
	mp.M_flags = M_FUNCTION;
	mp.M_narg = narg;
}


strmacro(name) {
	var	val  . BUFSIZE, p = 0;
	var	expb . BUFSIZE;
	var	ch, space, magic = 0;
	var	level = 1;

	ch = readsrc(Srcfile, 0);
	while (1) {
		if (#seof(Srcfile)) badeof();
		if (.ch == '(') {
			level += 1;
		}
		else if (.ch == ')') {
			if (!(level -= 1)) break;
		}
		if (magic && #isalpha(.ch) || .ch == '_')
			ch = expand(ch, expb, 0);
		if (space) {
			val.p = space; p += 1;
		}
		if (.ch == '#') {
			magic = 1;
		}
		else {
			if (p + strlen(ch) >= BUFSIZE) overflow();
			strcpy(val+p, ch);
			p += strlen(val+p);
			magic = 0;
		}
		ch = readsrc(Srcfile, @space);
	}
	newmac(name, val);
}


pp_define() {
	var	name, c;

	if (!State || Comment) return;

	if (skip() != '(') expect("(");
	name = getsymb();
	if ((c = skip()) == '(')
		fnmacro(name);
	else if (c == ',')
		strmacro(name);
	else
		expect(",");
}


pp_undef() {
	var	name, mac, macp;

	if (!State || Comment) return;

	if (skip() != '(') expect("(");
	name = getsymb();
	if (skip() != ')') expect(")");
	upcase(name);
	mac = Macros;
	macp = @Macros;
	while (mac) {
		if (!strcmp(name, mac.M_name)) {
			if (mac.M_flags & M_BUILTIN)
				report("will not #undef builtin macro");
			.macp = mac.M_next;
			mfree(mac.M_name);
			return;
		}
		macp = mac+M_next;
		mac = mac.M_next;
	}
	undefined(name);
}


defined() {
	var	name;

	if (skip() != '(') expect("(");
	name = getsymb();
	if (skip() != ')') expect(")");
	upcase(name);
	return findmac(name)-> TRUE, FALSE;
}


pushs(v) {
	if (Sptr >= STACKLEN) report("too many nested #if's");
	S_stack.Sptr = v; Sptr += 1;
}


pops() {
	if (!Sptr) nesting();
	return S_stack.(Sptr -= 1);
}


pp_if(exp) {
	if (Comment) return;
	if (State) {
		pushs(State);
		State = (exp == defined());
	}
	else {
		pushs(FALSE);
		defined();
	}
}


pp_ifdef() pp_if(TRUE);


pp_ifndef() pp_if(FALSE);


pp_else() {
	if (Comment) return;
	if (!Sptr) nesting();
	if (State) {
		State = FALSE;
	}
	else {
		if (S_stack.Sptr-1) State = TRUE;
	}
}


pp_endif() {
	if (Comment) return;
	State = pops();
}


pp_error() {
	var	msg, msgbuf . CHUNKSIZE;

	if (Comment || !State) return;
	if (skip() != '(') expect("(");
	msg = getstring(')');
	strfmt(msgbuf, "#error: %s", msg);
	report(msgbuf);
}


pp_comment() Comment += 1;


pp_endcomm() {
	if ((Comment -= 1) < 0)
		report("#cend without matching #comment");
}


ppmacro(name) {
	var	mac;

	upcase(name);
	if (!(mac = findmac(name))) {
		if (!State || Comment) return;
		undefined(name);
	}

	if (mac.M_flags & M_BUILTIN) {
		mac.M_val(); return;
	}

	if (!State || Comment) return;

	{
		var	expb . BUFSIZE;

		sputl(Outfile, expand(name, expb, 0));
	}
}


process() {
	var	chunk, magic = 0;

	Srcline = 1;
	swritef(Outfile, "# \"%s\" %ud%s", Srcname, Srcline, _NLCHAR_);

	chunk = readsrc(Srcfile, 0);
	while (chunk != ENDSTREAM) {
		if (magic && #isalpha(.chunk)) {
			ppmacro(chunk);
			magic = 0;
		}
		else {
			if (magic && State && !Comment)
				#sputc(Outfile, '#');
			magic = (.chunk == '#' && !chunk.1)-> 1, 0;
			if (State && !Comment && !magic)
				sputl(Outfile, chunk);
		}
		chunk = readsrc(Srcfile, 0);
	}
}


initdef(str) {
	var	v;

	if ((v = strchr(str, '='))) {
		.v = 0;
		v += 1;
	}
	else {
		v = 0;
	}
	newmac(str, v);
}


usage() {
	sputl(STDERR, "usage: bslpp [-p] [-i dir] [-dNAME[=VAL]] file\r\n");
	quit(1);
}


start(argcount, argvec) {
	var	arena  . POOLSIZE;
	var	argbuf . 128 / _BPW_;
	var	p;

	mminit(arena, POOLSIZE);
	stdopen();

	initmac();

	while (argcount) {
		unpack(.argvec, argbuf);
		if (.argbuf != '-') break;
		p = 1;
		while (argbuf.p) {
			switch (argbuf.p) {
				case 'p': case 'P':
					O_pipe = 1; break;
				
				case 'i': case 'I':
					if (argbuf.(p+1)) {
						Hdrpath = strsave(argbuf+p+1);
					}
					else {
						if (!(.(argvec += 1))) usage();
						argcount -= 1;
						unpack(.argvec, argbuf);
						Hdrpath = strsave(argbuf);
					}
					p += strlen(argbuf+p)-1;
					break;

				case 'd': case 'D':
					if (argbuf.(p+1)) {
						initdef(argbuf+p+1);
					}
					else {
						if (!(.(argvec += 1))) usage();
						argcount -= 1;
						unpack(.argvec, argbuf);
						initdef(argbuf);
					}
					p += strlen(argbuf+p)-1;
					break;

				default:
					usage();
			}
			p += 1;
		}
		argvec += 1;
		argcount -= 1;
	}

	if (argcount != 1) usage();

	Srcname = strsave(argbuf);
	if (O_pipe) {
		Srcfile = STDIN;
	}
	else {
		if ((Srcfile = sopen(Srcname, S_READ)) == ENDSTREAM) {
			swritef(STDERR, "bslpp: no such file: %s\r\n",
			Srcname);
			quit(1);
		}
	}

	Outfile = STDOUT;

	process();

	if (!O_pipe) sclose(Srcfile);

	if (Comment) report("#comment missing #cend at EOF");
	if (Sptr) report("unterminated #ifdef at EOF");

	quit(0);
}
