///////////////////////////////////////////////////////////////////////////////
//
//	NDXGEN -- Generate indices for BCODE libraries
//	Copyright (C) 1995 Nils M. Holm
//
//	NDXGEN searches each file specified in the list file 'Modules'
//	NDXGEN reads a module list from STDIN, and searches each
//	specified module for PUB and EXT records. It creates one index
//	entry containing the public symbols, and one containing the
//	external references of each module. The index is written to
//	STDOUT. Reports and error messages will be written to STDERR.
//
//	USAGE: NDXGEN [[-]V]
//
//	-V	Verbose operation. More v's give more information.
//	-F	Keep full module path names. Default is to remove
//		the path, and only store the file name.
//
//	An index entry has the following format:
//
//	<type> <module><TAB><symbol> ...
//
//	<type>		The record type. P = public, X = external
//	<module>	The module containing the symbols or links
//	<symbol> ...	A space-separated list of symbol names
//
//	When a module has no external references, no X-record will
//	be generated. When there are no public symbols in a module,
//	no P-record will be generated, but this is a weird case,
//	anyway.
//

#include( blib.hdr)
#include( ../ios/ios.hdr)
#include( ../mem/mem.hdr)
#include( ../str/str.hdr)

#include( ../opt.hdr)


var	NPub = 0, NExt = 0;

var	O_verbose = 0,
	O_fullpath = 0;


quit(n) {
	stdclose();
	halt n;
}


readword(src) {
	value	WORDLEN = 20;
	static var	buf.WORDLEN;
	var	c, i = 0;

	if (#seof(src)) return ENDSTREAM;

	c = #sgetc(src);
	while (#iswhite(c)) c = #sgetc(src);
	while (!#iswhite(c) && !#seof(src)) {
		buf.i = c; i += 1;
		if (i >= WORDLEN) {
			sputl(STDERR, "ndxgen: token too long\r\n");
			quit(1);
		}
		c = #sgetc(src);
	}
	buf.i = 0;

	return buf;
}


basename(path) {
	var	p;

	if ((p = strrchr(path, '/'))) return p+1;
	if ((p = strrchr(path, '\\'))) return p+1;
	return path;
}


writendx(module, type, src, dest) {
	var	nid, n = 0, fn;

	if (!(nid = stob(readword(src), 10))) return;

	if (type == 'P')
		NPub += nid;
	else
		NExt += nid;

	swritef(dest, "%c %s", type, O_fullpath-> module, basename(module));

	while (nid) {
		if ((fn = readword(src)) == ENDSTREAM) {
			sputl("ndxgen: unexpected EOF\r\n");
			quit(1);
		}
		swritef(dest, "%c%s", n-> '\s', '\t', fn);
		readword(src);
		nid -= 1;
		n += 1;
	}
	sputl(dest, _NLCHAR_);
}


append(name, src, dest) {
	var	word;

	while ((word = readword(src)) != ENDSTREAM) {
		if (!strcmp(word, "EXT"))
			writendx(name, 'X', src, dest);
		else if (!strcmp(word, "PUB"))
			writendx(name, 'P', src, dest);
	}
}


usage() {
	sputl(STDERR, "usage: ndxgen [[-]v]\r\n");
	quit(1);
}


start(argcount, argvec) {
	var	arena . 2048;
	var	buf . _PATHL_;
	var	argbuf . 128;
	var	bco;

	mminit(arena, 2048);
	stdopen();

	while (argcount) {
		unpack(.argvec, argbuf);
		.argvec = argbuf;
		if (..argvec == '-') .argvec += 1;
		while (..argvec) {
			switch (..argvec) {
				case 'v': case 'V':
					O_verbose += 1; break;

				case 'f': case 'F':
					O_fullpath += 1; break;

				default:
					usage();
			}
			.argvec += 1;
		}
		argvec += 1;
		argcount -= 1;
	}

	sgetl(STDIN, buf, _PATHL_);
	while (!#seof(STDIN)) {
		buf.(strlen(buf)-_NLSIZE_) = 0;
		if ((bco = sopen(buf, S_READ)) == ENDSTREAM) {
			swritef(STDERR, "ndxgen: no such file: %s\r\n", buf);
			ssync(STDERR);
		}
		else {
			if (O_verbose > 1) {
				swritef(STDERR, "+ %s\r\n", buf);
				ssync(STDERR);
			}
			append(buf, bco, STDOUT);
			sclose(bco);
		}
		sgetl(STDIN, buf, _PATHL_);
	}

	if (O_verbose) {
		swritef(STDERR, "= %ud symbols, %ud externals\r\n",
		NPub, NExt);
	}
	quit(0);
}
